-- Base script for SA2:Blast abilities

--GS: Made a quick table to check if you're one of the characters that should be affected by things.
if SA2BLASTSKINS == nil then rawset(_G, "SA2BLASTSKINS", {}) end
SA2BLASTSKINS["sonic"] = true
SA2BLASTSKINS["tails"] = true
SA2BLASTSKINS["knuckles"] = true
SA2BLASTSKINS["amy"] = true
SA2BLASTSKINS["fang"] = true
SA2BLASTSKINS["metalsonic"] = true
SA2BLASTSKINS["shadow"] = true
SA2BLASTSKINS["eggman"] = true
SA2BLASTSKINS["rouge"] = true

-- Handle button down inputs
local function sa2ButtonDown(p)
	if not (p and p.valid and p.playerstate == PST_LIVE) then return end -- Valid player?
	if p.cmd.buttons & BT_JUMP then p.sa2jumpdown = $ + 1 else p.sa2jumpdown = 0 end
	if p.cmd.buttons & BT_USE then p.sa2usedown = $ + 1 else p.sa2usedown = 0 end
end

addHook("PlayerThink", sa2ButtonDown)

-- Enable SA2 abilities for each character
local function sa2Abilities(p)
	if not (p and p.valid) then return end
	if p.sa2noabilities then return end
	
	-- Abilities for Sonic
	-- p.charability == CA_THOK then
	if p.mo.skin == "sonic" then
		p.charability = CA_HOMINGTHOK -- You are now evil.
		
		-- Homing attack is now a ball
		if (p.pflags & PF_THOKKED) and p.mo.state == S_PLAY_FALL
		and not (p.pflags & PF_JUMPED) then
			p.pflags = $ | PF_JUMPED
			p.mo.state = S_PLAY_JUMP
		end
		
		-- Thok keeps half momentum
		if (p.pflags & PF_THOKKED) and (p.pflags & PF_JUMPED)
		and p.sa2thokstart != nil then
			p.sa2thokstart = $ + 1
		else
			p.sa2thokstart = 0
		end
		
		if p.sa2thokstart == 1 then
			local xyvector = p.sa2thokspeed / 2 -- How fast were we going?
			local xy_mom = FixedHypot(p.mo.momx, p.mo.momy) -- How fast are we going now?
			if xyvector > xy_mom then
				local thokratio = FixedDiv(xyvector, xy_mom)
				p.mo.momx = FixedMul($, thokratio)
				p.mo.momy = FixedMul($, thokratio)
			end
		else
			p.sa2thokspeed = p.speed
		end
	end
	
	-- Abilities for Shadow
	if p.mo.skin == "shadow" then
		-- Nerf Chaos Snap count to 1 (Like in 2.1 SRB2 Heroes!)
		if p.shadow and p.shadow.snaps and
		(p.shadow.snaps > 1) or p.mo.eflags & MFE_SPRUNG then
			p.shadow.snaps = 1
		end
	end
	
	-- Mid-air abilties
	if not P_IsObjectOnGround(p.mo) and not (p.pflags & PF_THOKKED)
	and not p.powers[pw_carry] then
		p.pflags = $ | PF_JUMPED
	end

	-- Abilities for Knuckles and Rouge
	-- if p.charability == CA_GLIDEANDCLIMB then
	if p.mo.skin == "knuckles" or p.mo.skin == "rouge" then
	
		-- Multi-glide and aerial-glide
		if not P_IsObjectOnGround(p.mo) then
			p.pflags = $ | PF_JUMPED
			p.pflags = $ & ~PF_THOKKED
			p.charflags = $ | SF_MULTIABILITY
		end
		
		-- Punch ability! (Actually it's just Amy hammer but whatever)
		p.charability2 = CA2_MELEE
		p.powers[pw_strong] = $ & ~STR_SPRING
		p.mindash = 2 * FRACUNIT
		p.maxdash = 6 * FRACUNIT
		p.revitem = MT_LHRT
		
		-- Dumb hack to make the landing animation work properly
		if p.mo.state == S_PLAY_MELEE_LANDING and not p.sa2punch then
			p.sa2punch = leveltime
		end
		if p.sa2punch then
			if leveltime > p.sa2punch then
				p.mo.state = S_PLAY_RUN
				p.mo.frame = 1
				p.sa2punch = nil
			end
		end
	end
	
	-- Abilities for Tails and Eggman
	-- if p.charability == CA_FLY -- and P_IsObjectOnGround(p.mo) then
	if p.mo.skin == "tails" or p.mo.skin == "eggman" then
		p.charability = CA_SLOWFALL
		p.charflags = $ | SF_MULTIABILITY
		p.charability2 = CA2_GUNSLINGER
		p.revitem = MT_TORPEDO
		
	/*
	elseif p.charability == CA_FLOAT and
		   p.charflags & SF_MULTIABILITY and
	       p.charability2 == CA2_GUNSLINGER and
		   p.revitem == MT_TORPEDO then
	*/
		
		-- Enable hover in mid-air
		if not P_IsObjectOnGround(p.mo) and not (p.cmd.buttons & BT_JUMP) then
			p.pflags = $ | PF_JUMPED
			-- p.pflags = $ & ~PF_THOKKED
			-- p.charflags = $ | SF_MULTIABILITY
		end
		
		-- Logic for hover physics
		if (p.pflags & PF_THOKKED) and not (p.pflags & PF_JUMPED) then
			p.mo.state = S_PLAY_FLY
			
			if (P_MobjFlip(p.mo) > 0 and p.mo.momz > -2 * FRACUNIT)
			or (P_MobjFlip(p.mo) < 0 and p.mo.momz < 2 * FRACUNIT) then
				P_SetObjectMomZ(p.mo, -1 * FRACUNIT / TICRATE, true)
			else
				P_SetObjectMomZ(p.mo, -2 * FRACUNIT)
			end
			
			-- P_SetObjectMomZ(p.mo, 4 * FRACUNIT) -- 0.25 gravity for Cosmic Wall
			
			-- Restore flight sound effects
			if p.mo.skin == "tails" then
				if (leveltime % 10 == 1) then
					S_StartSound(p.mo, sfx_putput)
				end
			end
			if p.mo.skin == "eggman" then
				if (leveltime % 7 == 1) then
					S_StartSound(p.mo, sfx_jet)
				end
			end
			
			/*
			-- Stupid fucking hack for limiting float time
			p.sa2flighttimer = $ - 1
			if p.sa2jumpdown and p.sa2jumpdown == 1 then
				p.sa2flighttimer = $ - (TICRATE / 2)
			end
			if p.sa2flighttimer <= 0 then
				p.charability = CA_FLY
				p.powers[pw_tailsfly] = 0
				P_ResetPlayer(p)
			end
			*/
		end
		
		/*
		if P_IsObjectOnGround(p.mo) then
			p.sa2flighttimer = (3 * TICRATE) + (TICRATE / 2)
		end
		*/
	end
	
	-- Reset jumpfactor to default
	if p.mo.skin == "knuckles" or p.mo.skin == "eggman" then
		if p.jumpfactor != FRACUNIT then
			p.jumpfactor = FRACUNIT
		end
	end
	
end

addHook("PlayerThink", sa2Abilities)

freeslot("SPR_MFST")

-- Reskin melee hearts to fists
local function punchHearts(mo)
	if mo and mo.valid
	and mo.target and mo.target.player
	and mo.target.skin and mo.target.skin != "amy"
	and mo.target.skin != "rouge" then
		if mo.frame < 4096 * FRACUNIT then
			mo.sprite = SPR_MFST
			mo.frame = 0
		end
	end
end

addHook("MobjThinker", punchHearts, MT_LHRT)

-- Double the speed of torpedos you shoot
local function doubleTorpedoSpeed(mo)
	if mo and mo.valid and mo.target and mo.target.player then
		local totalspeed = FixedHypot(FixedHypot(mo.momx, mo.momy), mo.momz)
		if totalspeed > 0 and totalspeed < FixedMul(24 * FRACUNIT, mo.target.scale) then
			mo.momx = $ * 2
			mo.momy = $ * 2
			mo.momz = $ * 2
		end
	end
end

addHook("MobjThinker", doubleTorpedoSpeed, MT_TORPEDO)

freeslot("MT_GUNHUNTER")

-- Make homing attacks bounce off of enemies (target-side)
local function homingReboundTarget(target, inflictor, source)
	if source and source.valid and source.player
	and ((source.player.charability == CA_HOMINGTHOK and source.target == target)
	or (source.skin and source.skin == "shadow"
	and (target.type == MT_GUNHUNTER or target.sa2monitor))
	and source.tracer == target) then
		source.player.sa2homingangle = source.angle
		source.player.sa2homingdeath = true
		if source.skin and source.skin == "shadow" then
			source.player.sa2shadowhunter = true
		end
	end
end

addHook("MobjDeath", homingReboundTarget)

-- Make homing attacks bounce off of springs (target-side)
local function homingReboundSpring(thing, tmthing)
	if tmthing and tmthing.valid and tmthing.player
	and tmthing.player.charability == CA_HOMINGTHOK
	and tmthing.target == thing then
		tmthing.player.sa2homingangle = tmthing.angle
	end
end

addHook("MobjCollide", homingReboundSpring, MT_YELLOWSPRING)
addHook("MobjCollide", homingReboundSpring, MT_REDSPRING)
addHook("MobjCollide", homingReboundSpring, MT_BLUESPRING)

-- Make homing attacks bounce off of enemies/springs (player-side)
local function homingReboundPlayer(p)
	if p and p.valid and p.sa2homingangle then
		local reboundspeed = 16
		p.mo.momx = cos(p.sa2homingangle) * reboundspeed
		p.mo.momy = sin(p.sa2homingangle) * reboundspeed
		
		if (p.mo.momz <= 0 and p.sa2homingdeath and (p.pflags & PF_JUMPED))
		or p.sa2shadowhunter then
			P_ResetPlayer(p)
			P_SetObjectMomZ(p.mo, 10 * FRACUNIT)
			p.pflags = $ | PF_JUMPED
			p.pflags = $ | PF_NOJUMPDAMAGE
			p.mo.state = S_PLAY_SPRING
			-- p.pflags = $ & ~PF_THOKKED
			-- CONS_Printf(p, "THIS IS A BUG THAT I HAVE NO IDEA HOW TO CONSISTENTLY REPRODUCE")
			-- CONS_Printf(p, "AND THEREFORE HAVE NO IDEA HOW TO FIX AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA")
		end
		
		p.sa2homingangle = nil
		p.sa2homingdeath = nil
		p.sa2shadowhunter = nil
	end
end

addHook("PlayerThink", homingReboundPlayer)

-- Are these two objects colliding on the z-axis?
local function zCollide(thing, tmthing)
	if tmthing.z == thing.z then
		return true
	elseif tmthing.z > thing.z then
		return (thing.z + thing.height) > tmthing.z
	elseif tmthing.z < thing.z then
		return (tmthing.z + tmthing.height) > thing.z
	end
end

-- NiGHTS bumpers reset Shadow's snaps
local function bumperSnaps(thing, tmthing)
	if thing and thing.valid and tmthing and tmthing.valid
	and tmthing.player and zCollide(thing, tmthing) then
		local p = tmthing.player
		if p.sa2noabilities then return end
		if p.shadow and p.shadow.snaps == 0 then
			p.shadow.snaps = 1
		end
	end
end

addHook("MobjCollide", bumperSnaps, MT_NIGHTSBUMPER)