-- Bounce script - Loosely based off of Sonic Lost World Bounce: https://mb.srb2.org/addons/mo-lost-world-bounce-for-2-2.1177/

freeslot("sfx_sa2bnc")
sfxinfo[sfx_sa2bnc].caption = "Bounce"
local bouncedownspeed = -35 * FRACUNIT
local bounceupspeed = 14 * FRACUNIT

-- Reset bounce values whenever necessary
local function resetBounce(p)
	if p and p.valid then
		p.sa2bouncevalue = nil
		p.sa2bouncesnaps = nil
	end
end

addHook("PlayerSpawn", resetBounce)

-- Check if player is in stasis
local function stasisCheck(mo)
	if mo and mo.valid and mo.player and mo.player.valid then
		local p = mo.player
		if (p.pflags & PF_STASIS) or (p.pflags & PF_JUMPSTASIS) then
			p.stasislasttic = true
		else
			p.stasislasttic = false
		end
	end
end

addHook("MobjThinker", stasisCheck, MT_PLAYER)

-- Main function for bounce ability
local function bounceMain(p)

	-- Sanity/character checks
	if p.sa2noabilities then return end
	if not (p and p.valid and (p.mo.skin == "sonic" or p.mo.skin == "shadow")) then
		return
	end
	
	-- Stuff that can interrupt bounce
	if (p.mo.eflags & MFE_SPRUNG) or (p.mo.eflags & MFE_GOOWATER)
	-- or (p.mo.GSgrinding or (p.mo.GSgrind and p.mo.GSgrind.grinding)) --You don't need to specify this. Rails set pw_carry to 3888.
	or (P_MobjFlip(p.mo) > 0 and p.mo.momz > 0)
	or (P_MobjFlip(p.mo) < 0 and p.mo.momz < 0)
	or p.powers[pw_carry] or P_PlayerInPain(p)
	or p.playerstate != PST_LIVE or p.stasislasttic then
		p.sa2bouncevalue = 0
	end
	
	-- 0 = Not bouncing
	-- 1 = Going downwards
	-- 2 = Just hit the ground
	-- 3 = Force bounce (used for crates)
	if p.sa2bouncevalue == nil or p.sa2bouncevalue == 2 then p.sa2bouncevalue = 0 end
	
	-- Patch for making super transformation override bounce
	if P_SuperReady(p) and p.sa2usedown == 0 then p.sa2usedown = 2 end
	
	-- Propel the player downwards if they press spin
	if p.pflags & PF_JUMPED and not (p.pflags & PF_THOKKED)
	and p.sa2usedown != nil and p.sa2usedown == 1 and not p.exiting
	and p.sa2bouncevalue == 0 and not P_SuperReady(p)
	and not p.powers[pw_carry] and not P_PlayerInPain(p)
	and not p.playerstate != PST_LIVE and not p.stasislasttic then
		P_SetObjectMomZ(p.mo, bouncedownspeed)
		p.pflags = $ | PF_THOKKED
		p.pflags = $ & ~PF_SPINNING
		p.sa2bouncevalue = 1
		
		-- Patch for Shadow, prevent mid-snap bouncing
		if p.shadow and p.shadow.snaps != nil then
			p.sa2bouncesnaps = p.shadow.snaps
			p.shadow.snaps = 0
		end
	end
	
	-- Propel the player upwards when they hit the floor
	if ((p.mo.eflags & MFE_JUSTHITFLOOR) and p.sa2bouncevalue == 1)
	or p.sa2bouncevalue == 3 then
		p.pflags = $ | PF_JUMPED
		p.pflags = $ & ~PF_THOKKED
		P_SetObjectMomZ(p.mo, bounceupspeed)
		S_StartSound(p.mo, sfx_sa2bnc)
		p.sa2bouncevalue = 2
		
		-- Patch for Shadow, re-allow snaps after bouncing
		if p.shadow and p.shadow.snaps != nil then
			p.shadow.snaps = p.sa2bouncesnaps
			p.sa2bouncesnaps = nil
		end
	end
	
	-- Set proper animation
	if p.sa2bouncevalue == 1 or p.sa2bouncevalue == 2 then
		p.mo.state = S_PLAY_ROLL
	end
	
	-- Bounce stretch effect
	if p.sa2bouncevalue == 1 then
		p.mo.spritexscale = (1 * FRACUNIT) / 2
		p.mo.spriteyscale = (2 * FRACUNIT) / 1
	elseif p.mo.spritexscale == (1 * FRACUNIT) / 2
	and p.mo.spriteyscale == (2 * FRACUNIT) / 1 then
		p.mo.spritexscale = FRACUNIT
		p.mo.spriteyscale = FRACUNIT
	end
	
	-- Cap horizontal momentum for bounce (But not on speedpads!)
	if p.sa2bouncevalue == 2
	and (not p.sa2speedpad or leveltime > p.sa2speedpad + 3) then
		local xyvector = bounceupspeed * 4
		local xy_mom = FixedHypot(p.mo.momx, p.mo.momy)
		if xy_mom > xyvector then
			local bounceratio = FixedDiv(xyvector, xy_mom)
			p.mo.momx = FixedMul($, bounceratio)
			p.mo.momy = FixedMul($, bounceratio)
		end
	end
end

addHook("PlayerThink", bounceMain)