//Shadow's Battle Ability by SMS Alfredo
//Updated pvp homing behavior and balance adjustments by Krabs

local SPECIAL_DURATION = TICRATE+7
local SPECIAL_STARTUP = 1
local SPECIAL_ENDLAG = 10

//Thinker for the Burst FX
local burstthinker = function(mobj)
	if mobj and mobj.valid and mobj.target and mobj.target.valid
		mobj.tracer = mobj.target
		mobj.flags2 = $&~MF2_DONTDRAW
		if mobj.target.health and mobj.target.skin == "shadow"
		and mobj.target.player and mobj.target.player.actionstate
			mobj.momx = 0 mobj.momy = 0 mobj.momz = 0
			if mobj.target.eflags&MFE_VERTICALFLIP
				mobj.eflags = $|MFE_VERTICALFLIP
				P_TeleportMove(mobj, mobj.target.x, mobj.target.y, mobj.target.z+mobj.target.height)
			else
				P_TeleportMove(mobj, mobj.target.x, mobj.target.y, mobj.target.z)
			end
		else
			P_KillMobj(mobj)
		end
	end
end
addHook("MobjThinker", burstthinker, MT_CHAOSBURST1)
addHook("MobjThinker", burstthinker, MT_CHAOSBURST2)
addHook("MobjThinker", burstthinker, MT_CHAOSBURST3)

// Lach: moving some stuff to some separate functions so I can hijack them for spear charging effects >:)
rawset(_G, "Shadow_GetHandPositionFromAngle", function(target, angle)
	if target.eflags & MFE_VERTICALFLIP
		return P_ReturnThrustX(nil, angle, target.radius*3/2)+target.x,
			P_ReturnThrustY(nil, angle, target.radius*3/2)+target.y,
			target.z+target.height
	else
		return P_ReturnThrustX(nil, angle, target.radius*3/2)+target.x,
			P_ReturnThrustY(nil, angle, target.radius*3/2)+target.y,
			target.z
	end
end)
rawset(_G, "Shadow_SpawnChaosNightmare", function(mo, state)
	local orbs = {
		P_SpawnPlayerMissile(mo, MT_CHAOSNIGHTMARE, MF2_LINKDRAW|MF2_DONTDRAW),
		P_SpawnPlayerMissile(mo, MT_CHAOSNIGHTMARE, MF2_AMBUSH|MF2_LINKDRAW|MF2_DONTDRAW)
	}
	
	for i, orb in ipairs(orbs)
		if orb.valid
			orb.tracer = orb.target
			if state
				orb.state = state
			end
		end
	end
end)

//Thinker for the Nightmare FX
addHook("MobjThinker", function(mobj)
	if not mobj.valid return end
	local target = mobj.target
	local player = target.valid and target.player
	if player
	and target.health and target.skin == "shadow"
	and (player.actionstate or target.state == S_SHADOW_SPEARCHARGE and player.cmd.buttons & BT_USE)
		mobj.flags2 = $ & ~MF2_DONTDRAW | (target.flags2 & MF2_DONTDRAW)
		mobj.eflags = $ & ~MFE_VERTICALFLIP | (target.eflags & MFE_VERTICALFLIP)
		mobj.scale = target.scale // Lach: this is REALLY important for linkdraw, don't forget it!!
		local angle = player.drawangle+ANGLE_90
		if mobj.flags2&MF2_AMBUSH then angle = $+ANGLE_180 end
		mobj.momx = 0 mobj.momy = 0 mobj.momz = 0
		P_TeleportMove(mobj, Shadow_GetHandPositionFromAngle(target, angle))
	else
		P_KillMobj(mobj)
	end
end, MT_CHAOSNIGHTMARE)

//For damaging in single player
addHook("PlayerCanDamage", function(player)
	if player.actionstate and player.actiontime
	and player.actiontime > SPECIAL_ENDLAG and player.actiontime < SPECIAL_DURATION-SPECIAL_STARTUP
		return true
	end
end)

freeslot("SKINCOLOR_SHADOWGLOW")

local shadowmovecancel = function(player, mo)
	if P_PlayerInPain(player) or player.panim >= PA_PAIN or not mo.health
		//nothing
	elseif player.pflags&PF_SPINNING
		mo.state = S_PLAY_ROLL
	elseif P_IsObjectOnGround(mo)
		mo.state = S_PLAY_STND
	else
		mo.state = S_PLAY_FALL
		player.pflags = $|PF_THOKKED
		player.shadow.flags = $ | SHF_SHIELDABILITY //SHF_SHIELDABILITY
	end
	player.actionstate = 0
	if player.powers[pw_super]
		//sorry nothing
	elseif player.powers[pw_shield] & SH_FIREFLOWER
		mo.color = SKINCOLOR_WHITE
	else
		mo.color = player.skincolor
	end
	player.shieldscale = skins[mo.skin].shieldscale
	mo.shadowonthemove = false
end

//The colors relating to the burst frames
local burstcolors = {
	SKINCOLOR_BLUEBELL,
	SKINCOLOR_BLUEBELL,
	SKINCOLOR_SHADOWGLOW,
	SKINCOLOR_ROSY,
	SKINCOLOR_SHADOWGLOW,
	SKINCOLOR_BLUEBELL,
	SKINCOLOR_SHADOWGLOW,
	SKINCOLOR_ROSY,
	SKINCOLOR_BLUEBELL,
	SKINCOLOR_ROSY
}

//Loading code from Krabs
local ScriptLoaded = false

addHook("ThinkFrame",function() if not(ScriptLoaded) and CBW_Battle 
	ScriptLoaded = true
	local B = CBW_Battle
	
	//Allow Shadow himself to deal damage
	//Note how he has zero defense, which makes him vurnerable to projectiles
	local ShadowPriority = function(player)
		if player.actionstate and player.actiontime
		and player.actiontime > SPECIAL_ENDLAG
			if player.actionstate > 1
				B.SetPriority(player,0,69,nil,0,0,"Chaos Nightmare")
			else
				B.SetPriority(player,0,69,nil,0,0,"Chaos Burst")
			end
		elseif player.homing and player.mo and player.mo.state == S_PLAY_JUMP//Attraction shot uses S_PLAY_ROLL, so we can distinguish between the two with this
			B.SetPriority(player,1,1,nil,1,1,"homing attack")// only 1 def, unlike an attraction shot which is 2 def
		end
	end
	
	local IsHoming = function(player)
		if not (player and player.valid and player.playerstate == PST_LIVE)
			or not player.mo
			or not player.shadow
			or not (player.homing or player.shadow.battlehoming)
			or player.mo.state == S_PLAY_ROLL//attraction shot
			or not player.mo.health
			return false
		end
		return true
	end
	
	local ShadowPreCollide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
		if IsHoming(plr[n1])
			plr[n1].shadmarker = true
		end
	end
	
	local ShadowPostCollide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
		if plr[n1] and plr[n1].shadmarker
			plr[n1].shadmarker = nil
		end
	end
	
	local ShadowCollide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
		if not IsHoming(plr[n1])
			return false
		end
		
		local shadow = plr[n1].shadow
		plr[n1].homing = 0
		shadow.battlehoming = 0
		
		local speedlimit = 32
		
		if (hurt == 1 and n1 == 1) or (hurt == -1 and n1 == 2)
			//Code mostly lifted from shadow-main
			local HOMING_GHOST_TICS = 8 // amount of time to display thrusting homing attack ghosts
			local HOMING_THRUST_SOUND = sfx_shadhm // sound to play when performing a thrusting homing attack
			
			if mo[n1].eflags & MFE_UNDERWATER
				P_SetObjectMomZ(mo[n1], FixedDiv(457*FRACUNIT,72*FRACUNIT), false)
			else
				P_SetObjectMomZ(mo[n1], 10*FRACUNIT, false)
			end
			if shadow.snaps < 1
				shadow.snaps = 1
			end
			
			if plr[n1].cmd.buttons & BT_JUMP
				P_InstaThrust(mo[n1], mo[n1].angle, FixedMul(plr[n1].actionspd / 8, mo[n1].scale))
				S_StartSound(mo[n1], HOMING_THRUST_SOUND)
				shadow.ghosttics = HOMING_GHOST_TICS
				
				if mo[n2].type == MT_SPARRINGDUMMY or (plr[n2] and plr[n2].playerstate == PST_LIVE)
					P_InstaThrust(mo[n2],angle[n2] + ANGLE_180, FixedDiv(3*mo[n1].scale, weight[n2]))
					if plr[n2]
						plr[n2].drawangle = angle[n2]
					end
				end
			else
				P_InstaThrust(mo[n1],angle[n1],4 * mo[n1].scale)
				
				if mo[n2].type == MT_SPARRINGDUMMY or (plr[n2] and plr[n2].playerstate == PST_LIVE)
					P_InstaThrust(mo[n2],angle[n2], FixedDiv(5*mo[n1].scale, weight[n2]))
				end
			end
			
			mo[n1].state = S_SHADOW_BOUNCE
			for i = 1, shadow.homingchain % (3)
				mo[n1].state = S_SHADOW_POSE
			end
			plr[n1].pflags = $ | PF_NOJUMPDAMAGE
			
			return true
		end
		
		//Unsuccessful hit (bumped)
		P_SetObjectMomZ(mo[n1], mo[n2].scale*10 / B.WaterFactor(mo[n1]), false)
		P_InstaThrust(mo[n1],angle[n1],5 * mo[n1].scale)
		
		P_Thrust(mo[n1],angle[n1],thrust[n1] / 12)//Shadow's homing attack is so fast, we have to cut the bounce momentum to 1/12th
		if not pain[n1] 
			
			S_StartSoundAtVolume(mo[n1], sfx_habump, 170)//homing fail sfx
			
			if not (plr[n2] and (plr[n2].homing or (plr[n2].shadow and plr[n2].shadow.homing)))
				P_Thrust(mo[n1],angle[n2]+ANGLE_180,min(speedlimit*mo[n1].scale,thrust2[n1] / 12))//Shadow's homing attack is so fast, we have to cut the bounce momentum to 1/12th
			end
			B.DoPlayerFlinch(plr[n1], 8, angle[n1], thrust[n1])
			plr[n1].panim = PA_FALL
			mo[n1].state = S_PLAY_FALL
		end
		
		if (plr[n2] and plr[n2].valid)
			and (
				plr[n2].playerstate != PST_LIVE
				or plr[n2].shadmarker
				or plr[n2].climbing
				or pain[n2]
			)
			return true//Don't thrust climbing or shadmarker or dead people
		end
		
		//Thrust the enemy
		//print(FixedDiv(9*mo[n1].scale, weight[n2]) / FRACUNIT)
		//print(FixedDiv(5*mo[n1].scale, weight[n2]) / FRACUNIT)
		P_Thrust(mo[n2],angle[n2], FixedDiv(10*mo[n1].scale, weight[n2]))
		if not P_IsObjectOnGround(mo[n2])
			P_SetObjectMomZ(mo[n2], FixedDiv(5*mo[n1].scale, weight[n2]), true)
		end
		
		if plr[n2]
		and plr[n2].valid
			B.DoPlayerFlinch(plr[n2],(thrust[n2] * 2)/(mo[n1].scale * 4), angle[n2], thrust[n2])
		end
		
		return true
	end
	
	local ShadowSpecial = function(mo,doaction,pressedaction)
		local player = mo.player

		if P_IsObjectOnGround(mo)
			player.actiontext = "Chaos Burst"
			player.actionrings = 10
		else
			player.actiontext = "Chaos Nightmare"
			player.actionrings = 10
		end
		player.action2text = nil
		
		if player.actionstate
			player.actionsuper = true
		elseif player.actionsuper
			player.actionsuper = false
		end
		
		//Start the move
		if player.actionstate == 0
			if doaction == 1
			and not player.homing
			and not player.shadow.battlehoming
			and not (player.shadow.flags&~SHF_SHIELDABILITY)
			and not player.shadow.bike
				if P_IsObjectOnGround(mo)
					player.actionstate = 1
					S_StartSound(mo, sfx_chbrst)
					player.pflags = $|PF_FULLSTASIS&~PF_JUMPED&~PF_SHIELDABILITY
					P_SpawnPlayerMissile(mo, MT_CHAOSBURST1, MF2_LINKDRAW|MF2_DONTDRAW)
					player.shadowsdumbburst = P_SpawnPlayerMissile(mo, MT_CHAOSBURST2, MF2_LINKDRAW|MF2_DONTDRAW)
					P_SpawnPlayerMissile(mo, MT_CHAOSBURST3, MF2_LINKDRAW|MF2_DONTDRAW)
				else
					mo.momz = 0
					player.actionstate = 2
					player.pflags = $|PF_JUMPSTASIS&~PF_JUMPED&~PF_SHIELDABILITY&~PF_SPINNING
					S_StartSound(mo, sfx_chnght)
				end
				player.canguard = false
				player.actiontime = SPECIAL_DURATION
				mo.shadowonthemove = true
				B.PayRings(player)
				B.ApplyCooldown(player,TICRATE*5/2)
			elseif mo.shadowonthemove
				shadowmovecancel(player, mo)
			end
			
			if player.shieldscale < skins[mo.skin].shieldscale//dumb fix for that one bug that I can't repro
				player.shieldscale = skins[mo.skin].shieldscale
			end
			
		//End the move
		elseif not player.actiontime
		or P_PlayerInPain(player)
		or player.tumble
		or player.powers[pw_carry]
		or player.shadow.bike
			shadowmovecancel(player, mo)
			
		//Do the move
		elseif player.actionstate
			local maxdist = FixedMul(RING_DIST/6, mo.scale) //Attack range
			
			if mo.frame&FF_FRAMEMASK >= 2 and mo.tics != -1
				mo.tics = -1
			end
			
			//Nightmare
			if player.actionstate > 1
				maxdist = $*4/5 //Nightmare has a smaller range than Burst
				mo.momz = 0
				mo.momx = FixedMul($, FRACUNIT*96/100)
				mo.momy = FixedMul($, FRACUNIT*96/100)
				if mo.state != S_SHADOW_CHAOSNIGHTMARE
					mo.state = S_SHADOW_CHAOSNIGHTMARE
				end
				if mo.frame&FF_FRAMEMASK == 1 and mo.tics > 2
					mo.tics = 2
				end
				if player.actiontime == TICRATE+3
					Shadow_SpawnChaosNightmare(mo)
				end
				
			//Burst
			else
				if mo.state != S_SHADOW_CHAOSBURST
					mo.state = S_SHADOW_CHAOSBURST
				end
				if player.shadowsdumbburst and player.shadowsdumbburst.valid
					if player.actiontime > SPECIAL_ENDLAG and player.actiontime < SPECIAL_DURATION-SPECIAL_STARTUP
						if (player.shadowsdumbburst.frame & FF_FRAMEMASK < 2)
							player.shadowsdumbburst.frame = $ + 1
						end
					end
					if (mo.frame&FF_FRAMEMASK == 1 or mo.frame&FF_FRAMEMASK == 2)
						if burstcolors[player.shadowsdumbburst.frame&FF_FRAMEMASK]
							mo.color = burstcolors[player.shadowsdumbburst.frame&FF_FRAMEMASK]
						else
							if player.powers[pw_super]
								//sorry nothing
							elseif player.powers[pw_shield] & SH_FIREFLOWER
								mo.color = SKINCOLOR_WHITE
							else
								mo.color = player.skincolor
							end
							mo.frame = $+1
						end
					end
				end
			end
			
			//Shrink your shield
			if player.actiontime > TICRATE+4
				player.shieldscale = $*2/3
			elseif player.actiontime == TICRATE+4
				player.shieldscale = 0
			end
			
			//Damage from afar
			if player.actiontime > SPECIAL_ENDLAG and player.actiontime < SPECIAL_DURATION-SPECIAL_STARTUP
				local thokmobj = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
				thokmobj.radius = maxdist

				searchBlockmap("objects", function(refmobj, foundmobj)
					if foundmobj.flags & MF_NOCLIPTHING or foundmobj.health <= 0
					or foundmobj == mo or foundmobj.flags2 & MF2_FRET or not P_CheckSight(mo, foundmobj)
					or ((twodlevel or mo.flags2 & MF2_TWOD) and abs(mo.y-foundmobj.y) > mo.radius)
					return end

					if foundmobj.player
						if foundmobj.player.spectator or foundmobj.player.panim == PA_PAIN or G_TagGametype()
						or foundmobj.player.powers[pw_invulnerability]
						or foundmobj.player.powers[pw_super]
						or foundmobj.player.powers[pw_flashing]
							return
						end
					elseif (!((foundmobj.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR) and (foundmobj.flags & MF_SHOOTABLE))) == !(foundmobj.flags2 & MF2_INVERTAIMABLE))
						return // not a valid target
					end
		
					local zdist = (mo.z + mo.height/2) - (foundmobj.z + foundmobj.height/2)
					local dist = P_AproxDistance(mo.x-foundmobj.x, mo.y-foundmobj.y)

					dist = P_AproxDistance(dist, zdist)
					if dist > maxdist
						return // out of range
					end
					
					if foundmobj.player and B.MyTeam(player, foundmobj.player)
						return // friendly fire
					end
					
					if P_DamageMobj(foundmobj,mo,mo)
						if foundmobj.health
							P_Thrust(foundmobj, R_PointToAngle2(mo.x,mo.y,foundmobj.x,foundmobj.y), 4*mo.scale)
							if B.ZLaunch
								B.ZLaunch(foundmobj, 3*mo.scale, true)
							end
						end
					end
				end, thokmobj)
	
				P_KillMobj(thokmobj)
			
			//Grow your shield back
			end
			
			if player.actiontime <= 7
				if player.shieldscale < skins[mo.skin].shieldscale
					player.shieldscale = $+(skins[mo.skin].shieldscale/7)
				elseif player.shieldscale > skins[mo.skin].shieldscale
					player.shieldscale = skins[mo.skin].shieldscale
				end
			end
			
			player.canguard = false
			if player.actionstate == 1
				player.pflags = $|PF_FULLSTASIS&~PF_JUMPED&~PF_SHIELDABILITY
			else
				player.pflags = $|PF_JUMPSTASIS&~PF_JUMPED&~PF_SHIELDABILITY
			end
			player.actiontime = $-1
		end
	end
	
	B.SkinVars["shadow"] = {
		weight = 100,
		special = ShadowSpecial,
		func_priority_ext = ShadowPriority,
		func_precollide = ShadowPreCollide,
		func_collide = ShadowCollide,
		func_postcollide = ShadowPostCollide
	}
	
	print("Shadow's Battle Mod ability was loaded!")
end end)