/*
Changeable effect colors
	by Lach
v2
Reusable!
*/

local FILENAME = "client/shadow/deez.dat"
local MAXSPLITSCREENPLAYERS = 2
local DEFAULTVALUE = string.rep("None\n", MAXSPLITSCREENPLAYERS)
local COMMANDNAME = "skatecolor"
local COMMANDDESCRIPTION = "Sets the color of Shadow's skate flames. Can be the name or number of a skincolor!"
local VARNAME = "shadow_skatecolor"

local function valid(mo)
	return mo and mo.valid
end

// return a skincolor based on the name read from the file
local function ReadColor(line)
	local file = io.openlocal(FILENAME)
	
	if not file // no file? use the default
		return 0
	end
	
	while line > 1
		line = $ - 1
		file:read()
	end
	
	local val = R_GetColorByName(file:read())
	
	file:close()
	
	return val or 0
end

// save the skincolor names to the file
local function WriteColor(val, line)
	local n = 0
	local text
	local file = io.openlocal(FILENAME)
	
	if file
		text = file:read("*a")
		file:close()
	else
		text = DEFAULTVALUE
	end
	
	line = $ or 0
	while line > 1
		line = $ - 1
		n = text:find('\n', n + 1)
	end
	
	local fin = text:find('\n', n + 1)
	
	file = io.openlocal(FILENAME, "w+")
	file:write(text:sub(1, n), val and skincolors[val].name or "None", text:sub(fin))
	file:close()
end

// set this player's effect color to the given skincolor, and save
local function SetColor(player, n)
	player[VARNAME] = n
	if player == consoleplayer or player == secondarydisplayplayer
		WriteColor(player[VARNAME], player == secondarydisplayplayer and 2 or 1)
	end
end

local function ChangeColor(player, ...)
	local silent
	local args = {...}
	local arg = ""
	
	for i, word in ipairs(args)
		word = $:upper()
		if word == "-SILENT"
			silent = true
		else
			arg = $ .. word .. " "
		end
	end
	
	arg = $:sub(1, $:len() - 1)
	
	if not arg:len()
		if not silent
			CONS_Printf(splitscreen and consoleplayer or player,
				COMMANDDESCRIPTION..
				"\nCurrently set to "..(player[VARNAME] and skincolors[player[VARNAME]].name or "default")..".")
		end
		return
	end
	
	if arg == "DEFAULT"
	or arg == "OFF"
		SetColor(player, 0)
		return
	end
	
	local n = tonumber(arg)
	if n ~= nil
	and n >= 0 and n <= #skincolors
		SetColor(player, n)
		return
	end
	
	local exists, color = pcall(function(name) return _G["SKINCOLOR_"..name] end, arg)
	if exists
		SetColor(player, color)
		return
	end
	
	color = R_GetColorByName(arg)
	if color
		SetColor(player, color)
		return
	end
	
	if not silent
		CONS_Printf(splitscreen and consoleplayer or player,
			"Not a valid color!")
	end
end

COM_AddCommand(COMMANDNAME..2, ChangeColor, COM_SPLITSCREEN)
COM_AddCommand(COMMANDNAME, ChangeColor)

local configplayers = {}

addHook("ThinkFrame", do
	if isserver ~= (consoleplayer == server) // the client is player 1 while the gamestate is being received, avoid sending commands during this time
		return
	end
	
	for i, player in ipairs({consoleplayer, secondarydisplayplayer})
		if not valid(configplayers[i])
		and valid(player)
			configplayers[i] = player
			if player[VARNAME] == nil
				local text = COMMANDNAME
				if i == 2
					text = $..2
				end
				text = $ .. " -silent " .. ReadColor(i)
				COM_BufInsertText(consoleplayer, text)
			end
		end
	end
end)
			