------ RESULTS FUNCTIONS -----

-- Function to initialize player values for scoring system
local function sbInitScoringValues(p)
	p.sb_levelscore = 0
	p.sb_timebonus = 0
	p.sb_totalscore = 0
	
	p.sb_rankletter = nil
	p.sb_rankcolor = nil
	p.sb_rankoffset = nil
	
	p.sb_calculated = false
	p.sb_intermission = 0
end

addHook("PlayerSpawn", sbInitScoringValues)

if gamestate == GS_LEVEL then
	for player in players.iterate do
		sbInitScoringValues(player)
	end
end

-- Calculate the current time bonus for the player (y = -250x + 100000)
local function sbCalculateTimeBonus(p)
	local elapsedsecs = p.realtime / TICRATE
	local finaltimebonus = 100000 - (250 * elapsedsecs)
	return finaltimebonus
end

rawset(_G, "sbCalculateTimeBonus", sbCalculateTimeBonus)

-- Calculate a player's score and set the appropriate player values
local function calculateScore(p)

	-- Calculate your total score
	p.sb_levelscore = p.recordscore
	p.sb_timebonus = sbCalculateTimeBonus(p)
	if p.sb_timebonus < 0 then p.sb_timebonus = 0 end -- holy fucking shit
	p.sb_totalscore = p.sb_levelscore + p.sb_timebonus
	
	-- Calculate your actual rank
	local rankgrades = {
		["A"] = { color = SKINCOLOR_MAGENTA, offset = 11 },
		["B"] = { color = SKINCOLOR_RED, offset = 10 },
		["C"] = { color = SKINCOLOR_YELLOW, offset = 9 },
		["D"] = { color = SKINCOLOR_EMERALD, offset = 10 },
		["E"] = { color = SKINCOLOR_BLUE, offset = 10 }
	}
	
	-- Get the rank requirements for each rank letter
	local arankreq = 150000 -- Lua.SBARank = 150000
	if mapheaderinfo[gamemap].sbarank then -- Get score from level header if specified
		arankreq = tonumber(mapheaderinfo[gamemap].sbarank)
	end
    local rankscores = {arankreq,
					   (arankreq * 3) / 4,
					   (arankreq * 2) / 4,
					   (arankreq * 1) / 4}
	
	-- Award the rank letter based on your score
	if p.sb_totalscore == nil then p.sb_rankletter = "E"
    elseif p.sb_totalscore >= rankscores[1] then p.sb_rankletter = "A"
    elseif p.sb_totalscore >= rankscores[2] then p.sb_rankletter = "B"
    elseif p.sb_totalscore >= rankscores[3] then p.sb_rankletter = "C"
    elseif p.sb_totalscore >= rankscores[4] then p.sb_rankletter = "D"
	else p.sb_rankletter = "E"
    end
	
	-- Set the HUD color and offset for the rank letter
	p.sb_rankcolor = rankgrades[p.sb_rankletter].color
	p.sb_rankoffset = rankgrades[p.sb_rankletter].offset
	
	-- We've finished calculating score for this level!
	p.sb_calculated = true
	
end

local SB_EXITLEVEL = false

-- Custom exitlevel command
local function sbExitLevel()
	if multiplayer then
		SB_EXITLEVEL = true
	else
		if not usedCheats then
			print("Cheats must be enabled to force a level exit in single player.")
		else
			SB_EXITLEVEL = true
		end
	end
end

COM_AddCommand("exitlevel", sbExitLevel, COM_ADMIN)

-- Reset custom exitlevel variable
local function sbELMapChange()
	SB_EXITLEVEL = false
end

addHook("MapChange", sbELMapChange)

-- Determine if all players are in an exiting state
local function allExiting(p)
	if SB_EXITLEVEL then return true end
    local allexitstate = true
    for p in players.iterate do -- Check each player for their exiting state
		allexitstate = allexitstate and (p.exiting > 0 or p.bot > 0)
	end
    return allexitstate
end

----- INTERMISSION CONSTANTS -----

local resultstart = 105 -- When do the results start?

local levelappeartime = resultstart + 16 -- What time does the level score appear?
local timeappeartime = resultstart + 45 -- What time does the time bonus appear?
local totalappeartime = resultstart + 87 -- What time does the total score appear?
local rankappeartime = resultstart + 130 -- What time does the rank appear?

local finalxcoord = 160 -- Where do the HUD elements settle?
local randomfinishtime1 = 13 -- After the main text appears, when do the numbers stop randomizing?
local randomfinishtime2 = 18 -- After the main text appears, when do the numbers stop randomizing?
local resultend = resultstart + 210 + 35 -- What time do the results end?

----- INTERMISSION FUNCTIONS -----

-- What to do when a player has finished a level
local function handleFinishState(p)
	if not (p and p.valid) then return end -- Is the player valid?
	if NoSBSystem() then return end -- Combo system is disabled

	-- Once an individual player has finished a level
	if ((p.pflags & PF_FINISHED) or p.exiting > 0) then
	
		-- Internally calculate their score, if not already calculated
		if not p.sb_calculated then calculateScore(p) end
		
		-- Freeze recordscore to prevent from gaining more points post-finish
		if p.sb_intermission < resultend then p.recordscore = p.sb_levelscore
		else p.recordscore = p.sb_totalscore end -- Set as total score when revealed
		
	end

	-- If everyone has finished the level, start the intermission
	if allExiting(p) then
		p.exiting = 9999 -- Freeze p.exiting so the intermission can override it
		p.sb_intermission = $ + 1 -- Continually increment intermission timer
		
		-- Styles support
		p.tallytimer = 455 + 2 - p.sb_intermission
		if p.tallytimer >= 455 then p.exiting = 2*TICRATE+10 end
		if p.tallytimer <= 2 then p.tallytimer = 2 end
		
		-- Failsafe code for invalid rank values
		if p.sb_intermission == 10 then
			local sb1 = p.sb_levelscore
			local sb2 = p.sb_timebonus
			local sb3 = p.sb_totalscore
			local sb4 = p.sb_rankletter
			
			local failedRank = false
			if sb1 == nil or sb1 > 999999990 then failedRank = true end
			if sb2 == nil or sb2 > 999999990 then failedRank = true end
			if sb3 == nil or sb3 > 999999990 then failedRank = true end
			if sb4 == nil then failedRank = true end
			if sb4 != "A" and sb4 != "B" and sb4 != "C"
			and sb4 != "D" and sb4 != "E" then failedRank = true end
			
			if failedRank then
				CONS_Printf(p, "p.sb_levelscore " .. tostring(sb1))
				CONS_Printf(p, "p.sb_timebonus " .. tostring(sb2))
				CONS_Printf(p, "p.sb_totalscore " .. tostring(sb3))
				CONS_Printf(p, "p.sb_rankletter " .. tostring(sb4))
				CONS_Printf(p, "An error occurred when calculating your score!")
				CONS_Printf(p, "Please contact the mod creators, explain what happened,")
				CONS_Printf(p, "and send them your log so we can try and fix this bug!")
			
				p.sb_levelscore = 0
				p.sb_timebonus = 0
				p.sb_totalscore = 0
				p.sb_rankletter = "E"
			end
		end
	end
	
end

addHook("PlayerThink", handleFinishState)