----- INTERMISSION CONSTANTS -----

local resultstart = 105 -- When do the results start?

local levelappeartime = resultstart + 16 -- What time does the level score appear?
local timeappeartime = resultstart + 45 -- What time does the time bonus appear?
local totalappeartime = resultstart + 87 -- What time does the total score appear?
local rankappeartime = resultstart + 130 -- What time does the rank appear?

local finalxcoord = 160 -- Where do the HUD elements settle?
local randomfinishtime1 = 13 -- After the main text appears, when do the numbers stop randomizing?
local randomfinishtime2 = 18 -- After the main text appears, when do the numbers stop randomizing?
local resultend = resultstart + 210 + 35 -- What time do the results end?

----- INTERMISSION FUNCTIONS -----

-- Checking if a jingle has interrupted the level clear music
addHook("MusicChange", function(oldmusic, newmusic)
    if NoSBSystem() then return end -- Combo system is disabled
    if oldmusic == "SBCLR" and newmusic != "SBCLR" then -- From SBCLR to something else
        for p in players.iterate do p.sb_jingleinterrupt = true end
    end
end)

-- Function to implement Jump/Spin/AttackDown
local function sbButtonDown(p)
	if NoSBSystem() then return end -- Combo system is disabled
	if p.cmd.buttons & BT_JUMP then p.sb_jumpdown = $ + 1 else p.sb_jumpdown = 0 end
	if p.cmd.buttons & BT_SPIN then p.sb_spindown = $ + 1 else p.sb_spindown = 0 end
end

addHook("PlayerThink", sbButtonDown)

local RETURN_STAGE = nil

-- Function for custom exiting level functionality
local function handleExitingLevel()
	token = 0 -- Disable token functionality for now

	-- Save the stage we were originally going to
	if token > 0 and not G_IsSpecialStage(gamemap) then
		RETURN_STAGE = mapheaderinfo[gamemap].nextlevel
	end
	
	local sbnextlevel = mapheaderinfo[gamemap].nextlevel or (gamemap + 1)
	
	-- Leaving the special stage rotation? Override nextlevel
	if G_IsSpecialStage(gamemap) and (token == 0 or All7Emeralds(emeralds)) then
		if RETURN_STAGE then
			sbnextlevel = RETURN_STAGE
		else
			sbnextlevel = spstage_start
		end
	end

	-- Change skin right before leaving so the save icon displays properly
	if not multiplayer and not token and mapheaderinfo[sbnextlevel]
	and mapheaderinfo[sbnextlevel].forcecharacter then
		local nextskin = mapheaderinfo[sbnextlevel].forcecharacter
		if nextskin != "" then R_SetPlayerSkin(players[0], nextskin) end
	end
	
	G_SetCustomExitVars(sbnextlevel, 1)
	G_ExitLevel()
end

-- Handling internal intermission logic
local function handleIntermission(p)
	if not (p and p.valid) then return end -- Is the player valid?
	if NoSBSystem() then return end -- Combo system is disabled
	
	-- Skip intermission in special stages
	if p.sb_intermission == resultstart - 1 then
		if G_IsSpecialStage(gamemap)
		or mapheaderinfo[gamemap].lvlttl == "Chao World" then
			handleExitingLevel()
		end
	end

	-- Only when the results have actually started displaying
	if p.sb_intermission > 0 -- >= resultstart then

		-- Halt movement during results
		if p.sb_intermission == 1 then
			P_ResetPlayer(p)
			p.mo.state = S_PLAY_WAIT
		end
	
		-- At what time will G_ExitLevel() be called?
		local exitleveltime = resultend + 105
		local sbinttime = CV_FindVar("inttime").value
		if multiplayer and sbinttime > 3 then -- Minimum of 3 second wait
			exitleveltime = resultend + (sbinttime * TICRATE)
		end

		-- Skip intermission with Spin (Only admins can skip in MP)
		if p.sb_intermission < exitleveltime - 105 + 1
		and p.sb_intermission > TICRATE * 2
		and p.sb_spindown == 1 then
			if (not multiplayer and not modeattacking)
			or (multiplayer and ((p == server) or IsPlayerAdmin(p))) then
				for player in players.iterate do
					if player.sb_intermission < resultend then
						S_StopMusic(player)
						-- S_StartSound(nil, sfx_doord2, player)
						P_AddPlayerScore(player, player.sb_timebonus)
					end
					player.sb_intermission = exitleveltime - 105 + 1
				end
			end
		end
	
		/*
		-- When the results screen is just starting, play the results music
		if p.sb_intermission == resultstart then
			S_ChangeMusic("SBCLR", false, p)
		end
		
		-- If the results music was interrupted by a jingle, force the results music back
		if p.sb_intermission >= resultstart and p.sb_jingleinterrupt then
			local ticposition = p.sb_intermission - resultstart
            local musposition = G_TicsToSeconds(ticposition) * 1000 + G_TicsToMilliseconds(ticposition)
            S_ChangeMusic("SBCLR", false, p, 0x4000, musposition)
            p.sb_jingleinterrupt = false
		end
		*/
		
		-- Add time bonus to actual score when results are done
		if p.sb_intermission == resultend - 50 then
			P_AddPlayerScore(p, p.sb_timebonus)
		end
		
		-- Exit the level, skipping the normal results screen
		if p.sb_intermission == exitleveltime then
			handleExitingLevel()
		end
	
	end
	
end

addHook("PlayerThink", handleIntermission)

/*
Draw results element based on the following information:
label        - The label of the element      
value        - The p.[value] to be drawn
x            - X coordinate of the label
y            - Y coordinate of the label
xoffset      - X offset of the value from the label
yoffset      - Y offset of the value from the label
movingxlabel - X coordinate of the label when moving
appeartime   - Time when this element starts appearing
randomtime   - Time after settling, when the value stops randomizing
*/
local function drawResultsElement(v, p, label, value, x, y, xoffset, yoffset,
							      movingxlabel, appeartime, randomtime)
	
	-- Draw text in their final positions
	if p.sb_intermission >= appeartime then
		
		-- Draw label
        v.drawString(x, y, label, V_PERPLAYER, "center")
		
		-- Draw value (Randomizing at first)
		if p.sb_intermission >= appeartime + randomtime then
			v.drawNum(x + xoffset, y + yoffset, p[value], V_PERPLAYER, "center")
		else
			v.drawNum(x + xoffset, y + yoffset, v.RandomRange(10000, 65536), V_PERPLAYER, "center")
		end
	
	-- Draw text in their moving positions
	elseif p.sb_intermission >= appeartime - 10 then
		v.drawString(movingxlabel, y, label, V_PERPLAYER, "center")
		v.drawNum((x * 2) + (xoffset - movingxlabel), y + yoffset, v.RandomRange(10000, 65536), V_PERPLAYER, "center")
	end
	
end

/*
Draw rank element based on the following information:
x            - X coordinate of the label
y            - Y coordinate of the label
xoffset      - X offset of the box from the label
yoffset      - Y offset of the box from the label
movingxlabel - X coordinate of the label when moving
appeartime   - Time when this element starts appearing
*/
local function drawRankElement(v, p, x, y, xoffset, yoffset, movingxlabel, appeartime)
	
	-- Moving coordinates for the rank elements
	local labelx = movingxlabel
	local outerx = (x * 2) + xoffset - movingxlabel
	local innerx = (x * 2) + xoffset + 2 - movingxlabel
	local letterx = (x * 2) - p.sb_rankoffset - movingxlabel
	
	-- Make elements static after they reach the destination
	if p.sb_intermission >= appeartime then
		labelx = x
		outerx = x - xoffset
		innerx = x - xoffset + 2
		letterx = x - p.sb_rankoffset
	end
	
	-- Draw the moving elements
	if p.sb_intermission >= appeartime - 10 then
        v.drawString(labelx, y, "RANK", V_PERPLAYER, "center")
		v.drawFill(outerx, y + yoffset, 40, 30, SKINCOLOR_WHITE) -- Rank Box Border
        v.drawFill(innerx, y + yoffset + 2, 40 - 4, 30 - 4, skincolors[p.sb_rankcolor].ramp[3]) -- Rank Box Inner
		v.drawNameTag(letterx, y + yoffset + 5, p.sb_rankletter, V_PERPLAYER, SKINCOLOR_WHITE, p.sb_rankcolor) -- Rank Letter
	end
	
end

/*
Draw rank patch based on the following information:
x            - X coordinate of the label
y            - Y coordinate of the label
xoffset      - X offset of the patch from the label
yoffset      - Y offset of the patch from the label
movingxlabel - X coordinate of the label when moving
appeartime   - Time when this element starts appearing
*/
local function drawRankPatch(v, p, x, y, xoffset, yoffset, movingxlabel, appeartime)
	
	-- Moving coordinates for the rank elements
	local labelx = movingxlabel
	local patchx = (x * 2) + (xoffset - movingxlabel)
	
	-- Make elements static after they reach the destination
	if p.sb_intermission >= appeartime then
		labelx = x
	    patchx = x - xoffset
	end 
	
	-- Draw the moving elements
	if p.sb_intermission >= appeartime - 10 then
		v.drawString(labelx, y, "RANK", V_PERPLAYER, "center")
		v.drawScaled(patchx * FRACUNIT, (y + yoffset) * FRACUNIT,
		             3 * FRACUNIT / 4 , v.cachePatch("RTRANK" .. p.sb_rankletter))
	end
	
end

-- Returns true if score_a is greater than score_b, used to sort score list
local function sortScores(score_a, score_b)
	return score_a[2] > score_b[2]
end

-- Gets the scores of all the players in the game, in descending order
local function getScoreList()
	local scorelist = {}
	
	-- Get the scores from all the players in the game
	for p in players.iterate do
		if p.bot == 0 and p.name != nil and p.sb_totalscore != nil then
			table.insert(scorelist, {p.name, p.sb_totalscore})
		end
	end
	
	
	-- Test score list
	local scorelist = {
		{"Alice", 85000}, {"Bob", 42000}, {"Carol", 64000}, {"David", 72000}, {"Emily", 31000},
		{"Frank", 54000}, {"Grace", 98000}, {"Henry", 74000}, {"Irene", 22000}, {"Jack", 89000}
	}
	
	
	-- Sort the score list in descending order
	table.sort(scorelist, sortScores)
	return scorelist
end

-- Stupid way of dynamically updating the transparency for the results box
local function renderResultsBox(v, p)
	local translist = { V_90TRANS, V_80TRANS, V_70TRANS, V_60TRANS,
						V_50TRANS, V_40TRANS, V_30TRANS, V_20TRANS }
	local transchoose = translist[8]
	
	if p.sb_intermission < resultstart + 6 then
		transchoose = translist[p.sb_intermission - resultstart + 1]
	end
	
	v.drawScaled(105 * FRACUNIT, 15 * FRACUNIT, FRACUNIT / 4, v.cachePatch("SCBX"), transchoose)
end

-- Render the elements for the results screen
local function renderResultsScreen(v, p)
	if NoSBSystem() then return end -- Combo system is disabled

	-- Only when the results are actually supposed to start displaying
	if p.sb_intermission >= resultstart then
	
		-- Fade in the results box behind everything
		-- renderResultsBox(v, p)
		
		-- Moving coordinates for the labels
		local levellabelx = ((-32 * ((p.sb_intermission - levelappeartime) ^ 2)) / 5) + finalxcoord
		local timelabelx = ((-32 * ((p.sb_intermission - timeappeartime) ^ 2)) / 5) + finalxcoord
		local totallabelx = ((-32 * ((p.sb_intermission - totalappeartime) ^ 2)) / 5) + finalxcoord
		local ranklabelx = ((-32 * ((p.sb_intermission - rankappeartime + 50) ^ 2)) / 5) + finalxcoord
		
		/*
		-- Draw each HUD element
		drawResultsElement(v, p, "LEVEL SCORE", "sb_levelscore", finalxcoord, 25, 20, 15,
		                   levellabelx, levelappeartime, randomfinishtime1)
		drawResultsElement(v, p, "TIME BONUS", "sb_timebonus", finalxcoord, 60, 20, 15,
		                   timelabelx, timeappeartime, randomfinishtime2)
		drawResultsElement(v, p, "TOTAL SCORE", "sb_totalscore", finalxcoord, 95, 20, 15,
		                   totallabelx, totalappeartime, rankappeartime - 3 - totalappeartime)
		drawRankElement(v, p, finalxcoord, 130, 20, 15, ranklabelx, rankappeartime)
		-- drawRankPatch(v, p, finalxcoord, 130, 15, 15, ranklabelx, rankappeartime)
		*/
		
		-- Draw intermission timer (Note: This has a minimum of 3 seconds)
		if multiplayer and p.sb_intermission > resultend then
			local rtinttime = CV_FindVar("inttime").value
			if rtinttime < 3 then rtinttime = 3 end
			local timeleft = (resultend + (rtinttime * TICRATE)) - p.sb_intermission
			v.drawString(160, 188,
			             "START IN " .. (timeleft + (TICRATE - 1)) / TICRATE .. " SECONDS",
						 V_PERPLAYER | V_YELLOWMAP, "center")
		end
		
		-- Draw leaderboard elements in multiplayer (DISABLED FOR NOW)
		if p.sb_intermission >= rankappeartime - 10 + 50 and false then -- and multiplayer then

			-- Get moving X coordinates for leaderboard elements
			local boardcoord1 = ranklabelx - (finalxcoord - 40)
			local boardcoord2 = ranklabelx - (finalxcoord - 75)
			local boardcoord3 = (280 * 2) - (ranklabelx - (finalxcoord - 280))
			local boardcoord4 = (250 * 2) - (ranklabelx - (finalxcoord - 250))
			if p.sb_intermission >= rankappeartime + 50 then
				boardcoord1 = 40
				boardcoord2 = 75
				boardcoord3 = 280
			    boardcoord4 = 250
			end
			
			-- Get the scores of all the players in the game!
			local scorelist = getScoreList()

			-- Top three place markers
			local flags1 = V_PERPLAYER | V_SNAPTOLEFT | V_CENTERNAMETAG

			if #scorelist >= 1 then v.drawNameTag(boardcoord1, 60, "1st", flags1, SKINCOLOR_GOLD, SKINCOLOR_GOLD) end
			if #scorelist >= 2 then v.drawNameTag(boardcoord1, 100, "2nd", flags1, SKINCOLOR_SILVER, SKINCOLOR_SILVER) end
			if #scorelist >= 3 then v.drawNameTag(boardcoord1, 140, "3rd", flags1, SKINCOLOR_BRONZE, SKINCOLOR_BRONZE) end
			
			-- Top three players and scores
			local flags2 = V_YELLOWMAP | V_PERPLAYER | V_SNAPTOLEFT | V_ALLOWLOWERCASE
			
			if #scorelist >= 1 then
				v.drawString(boardcoord2, 60 + 0, scorelist[1][1], flags2, "left")
				v.drawString(boardcoord2, 60 + 10, scorelist[1][2], flags2, "left")
			end
			
			if #scorelist >= 2 then
				v.drawString(boardcoord2, 100 + 0, scorelist[2][1], flags2, "left")
				v.drawString(boardcoord2, 100 + 10, scorelist[2][2], flags2, "left")
			end
			
			if #scorelist >= 3 then
				v.drawString(boardcoord2, 140 + 0, scorelist[3][1], flags2, "left")
				v.drawString(boardcoord2, 140 + 10, scorelist[3][2], flags2, "left")
			end

			-- Next five place markers
			local flags3 = V_PERPLAYER | V_SNAPTORIGHT | V_ALLOWLOWERCASE
			
			if #scorelist >= 4 then v.drawString(boardcoord3, 95 - 50, "4th", flags3, "right") end
			if #scorelist >= 5 then v.drawString(boardcoord3, 95 - 25, "5th", flags3, "right") end
			if #scorelist >= 6 then v.drawString(boardcoord3, 95, "6th", flags3, "right") end
			if #scorelist >= 7 then v.drawString(boardcoord3, 95 + 25, "7th", flags3, "right") end
			if #scorelist >= 8 then v.drawString(boardcoord3, 95 + 50, "8th", flags3, "right") end
			
			-- Next five players and scores
			local flags4 = V_YELLOWMAP | V_PERPLAYER | V_SNAPTORIGHT | V_ALLOWLOWERCASE
			
			if #scorelist >= 4 then
				v.drawString(boardcoord4, 95 - 50 - 5, scorelist[4][1], flags4, "right")
				v.drawString(boardcoord4, 95 - 50 + 5, scorelist[4][2], flags4, "right")
			end

			if #scorelist >= 5 then
				v.drawString(boardcoord4, 95 - 25 - 5, scorelist[5][1], flags4, "right")
				v.drawString(boardcoord4, 95 - 25 + 5, scorelist[5][2], flags4, "right")
			end

			if #scorelist >= 6 then
				v.drawString(boardcoord4, 95 - 5, scorelist[6][1], flags4, "right")
				v.drawString(boardcoord4, 95 + 5, scorelist[6][2], flags4, "right")
			end
			
			if #scorelist >= 7 then
				v.drawString(boardcoord4, 95 + 25 - 5, scorelist[7][1], flags4, "right")
				v.drawString(boardcoord4, 95 + 25 + 5, scorelist[7][2], flags4, "right")
			end
			
			if #scorelist >= 8 then
				v.drawString(boardcoord4, 95 + 50 - 5, scorelist[8][1], flags4, "right")
				v.drawString(boardcoord4, 95 + 50 + 5, scorelist[8][2], flags4, "right")
			end
			
		end
	end
end

hud.add(renderResultsScreen)

-- More detailed score info
local function extraSBInfo(v, p)
	if NoSBSystem() then return end -- Combo system is disabled

	local earnflags = V_SNAPTOTOP | V_SNAPTORIGHT | V_ALLOWLOWERCASE
	v.drawString(320 - 8, 34, sbCalculateTimeBonus(p), earnflags, "right")
	v.drawString(320 - 8, 45, sbCalculateTimeBonus(p) + p.recordscore, earnflags, "right")
end

-- hud.add(extraSBInfo)

-- Stop tally music from returning a second time
local function stopTallyMusicReturn(oldname, newname)
	if newname == "_ADVCL" and consoleplayer.sb_intermission > rankappeartime then
		return true
	end
end

addHook("MusicChange", stopTallyMusicReturn)