----- COMBO SYSTEM CONSTANTS -----

local sb_combolength = 5 * TICRATE -- How long should the combo last?
local sb_milestone = 1000 -- How many points do you need for a bonus?

----- MAIN COMBO SYSTEM FUNCTIONS -----

-- Initialize necessary player variables for combo system (resets after every combo)
local function initComboVars(p)
    p.sb_combotimer = -1 -- Time left until the combo ends
	p.sb_comboactive = false -- Is the combo currently active?
	p.sb_comboframe = 1 -- What star frame are we currently displaying?
	
    p.sb_startscore = p.score -- Your score when the combo started
    p.sb_highscore = p.score -- The highest score you've reached during this combo
	p.sb_prevscore = p.score -- Your score on the previous tic
	
	p.sb_lastcomboscore = 0 -- Information on your last earned combo score
	p.sb_lastcombomulti = 0 -- Information on your last earned combo multiplier
	p.sb_lastcomboend = -9999 -- What time did your last combo end?
end

addHook("PlayerSpawn", initComboVars)

-- Initialize necessary player variables for combo system (only on actual spawn)
local function loadComboVars(p)
	p.sb_lastcarrytype = 0 -- The PW_CARRY type the player had on the previous frame
	p.sb_carrystarttime = -9999 -- The time when the player started getting carried
	p.sb_lastsprungtime = -9999 -- The time when the player was last sprung
	
	p.sb_grindstarttime = -9999 -- The time when the player started grinding
	p.sb_grindlastframe = 0 -- Were we grinding on the last frame?
	
	p.sb_slidelastframe = 0 -- Were we in a waterslide on the last frame?
	p.sb_slidestarttime = -9999 -- When did we start watersliding?
	p.sb_touchedobjects = {} -- Objects that I have shoved up my a
	
	p.sb_finishlastframe = false -- Did we finish the level on the last frame?
	p.sb_prevscore2 = p.score -- Your score on the previous tic
	
	p.sb_springscore = 200 * 20 -- How much score can we earn from springs?
	p.sb_continuousscore = 10 * 60 -- How much score can we earn from continuous actions?
	p.sb_grindscore = 10 * 120 -- How much score can we earn from grinding?

	p.sb_shieldprevtic = 0 -- Did we have a shield on the previous tic?
end

addHook("PlayerSpawn", loadComboVars)

if gamestate == GS_LEVEL then
	for player in players.iterate do
		initComboVars(player)
		loadComboVars(player)
	end
end

-- Freeze the player's scoreAdd at 1
local function noScoreAdd(target, inflictor, source)
	if NoSBSystem() then return end -- Combo system is disabled
    if source and source.player and source.player.valid then
		source.player.scoreadd = 1 -- All enemies give 200 points!
	end
end

addHook("MobjDamage", noScoreAdd)

/*
-- Intervals for each score multiplier
-- This is y = 200 * (2 ^ ((x + 1) / 2))
-- Rounded to the nearest 200
local comboIntervals = {
    400, 600, 800, 1200, 1600, 
	2200, 3200, 4600, 6400, 9000
}

-- Function to determine your combo multiplier
local function findComboMultiplier(comboscore)
    for i = 10, 1, -1 do
        if comboscore >= comboIntervals[i] then
            return i
        end
    end
    return 0
end

-- What happens when the combo timer runs out
local function finishCombo(p)
	-- Give the player their bonus points!
	local comboscore = p.sb_highscore - p.sb_startscore
    local combomultiplier = findComboMultiplier(comboscore)
    P_AddPlayerScore(p, (comboscore * combomultiplier) / 2)
    initComboVars(p) -- And reset all relevant values for the next combo
	
	-- Save relevant combo info
    p.sb_lastcomboscore = comboscore
	sb_lastcombomulti = combomultiplier
	p.sb_lastcomboend = p.jointime
end
*/

-- Main function to handle the combo system!
local function comboSystemMain(p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid) then return end

	-- Finishing the level awards 1000 points
	if p.pflags & PF_FINISHED then
		if not p.sb_finishlastframe then
			if not mapheaderinfo[gamemap].sa2ecount then
				P_AddPlayerScore(p, 1000)
			end
		end
	end

	-- If the player doesn't have an active combo
	if not p.sb_comboactive then
		-- Combo timer starts if you get any points
        if p.score > p.sb_prevscore and not p.exiting then
			p.sb_comboactive = true
            p.sb_combotimer = sb_combolength
            p.sb_startscore = p.sb_prevscore
			p.sb_comboframe = 1
        end
	end
	p.sb_prevscore = p.score
    
	-- If the player has an active combo
	local finishstate = p.pflags & PF_FINISHED and not p.sb_finishlastframe
	if p.sb_comboactive or finishstate then
	
		-- Tick down the combo clock
		p.sb_combotimer = $ - 1 
		
		/*
		-- Combo instantly ends if you get hurt without a shield
        if (P_PlayerInPain(p) and not p.sb_shieldprevtic)
		or p.playerstate != PST_LIVE then
			initComboVars(p)
		end
		if not P_PlayerInPain(p) then
			p.sb_shieldprevtic = p.powers[pw_shield]
		end
		*/
        
        -- Combo timer resets if you get more points mid-combo
        if p.score > p.sb_highscore then
            p.sb_highscore = p.score
            p.sb_combotimer = sb_combolength
        end
		
		-- Finishing the level instantly ends your combo
		if p.pflags & PF_FINISHED then
			if not p.sb_finishlastframe then
				p.sb_combotimer = 0
			end
		end
		
		-- Check for when you reach a 1k milestone
		local scoreinterval = 0
		local scorecrossed = false
		if p.score > p.sb_prevscore2 then
			for i = 1, 100 do
				scoreinterval = p.sb_startscore + (i * sb_milestone)
				if p.score >= scoreinterval
				and scoreinterval > p.sb_prevscore2 then
					-- CONS_Printf(p, "You crossed " .. tostring(scoreinterval))
					local multiplier = i % 10
					if multiplier == 0 then multiplier = 10 end
					
					-- Insert interval functionality here
					if not scorecrossed then
						scorecrossed = true
						p.sb_lastcomboscore = multiplier * sb_milestone
						p.sb_lastcombomulti = multiplier
						p.sb_lastcomboend = p.jointime
					elseif scorecrossed
						scorecrossed = true
						p.sb_lastcomboscore = $ + (multiplier * sb_milestone)
						if multiplier > p.sb_lastcombomulti then
							p.sb_lastcombomulti = multiplier
						end
						p.sb_lastcomboend = p.jointime
					end
				end
			end
		end
		
		-- STUPID FUCKING HACK I HATE THIS SO MUCH
		local savelcs, savelcm, savelce
		if scorecrossed then
			P_AddPlayerScore(p, p.sb_lastcomboscore)
			p.sb_startscore = $ + p.sb_lastcomboscore
			savelcs = p.sb_lastcomboscore
			savelcm = p.sb_lastcombomulti
			savelce = p.sb_lastcomboend
		end
		p.sb_prevscore2 = p.score
        
		-- If the combo timer runs out
		if p.sb_combotimer == 0 then
			-- local scoreoffset = sb_milestone - (p.sb_highscore % sb_milestone)
			-- local new_startscore = p.sb_startscore + 3000 - scoreoffset
			-- if new_startscore > p.score then new_startscore = p.score end
            initComboVars(p)
			-- p.sb_startscore = new_startscore
			if scorecrossed and p.exiting then
				p.sb_lastcomboscore = savelcs
				p.sb_lastcombomulti = savelcm
				p.sb_lastcomboend = savelce
			end
			-- finishCombo(p)
        end
    end
	
	p.sb_finishlastframe = p.pflags & PF_FINISHED -- Update finish state
	
end

addHook("PlayerThink", comboSystemMain)

----- AWARD POINTS FOR DOING STUFF -----

-- Earn points for interacting with this object! But only once, though
local function earnPointsForObject(p, mo, points)
	if p and p.valid and mo and mo.valid then
		if not p.sb_touchedobjects[mo] then
			P_AddPlayerScore(p, points)
			p.sb_touchedobjects[mo] = true
		end
	end
end

-- Award points for collecting items like rings and monitors
-- Note: By default, enemies give 200 points, and tokens give 1000 points
local function awardItemPoints(target, inflictor, source)
	if NoSBSystem() then return end -- Combo system is disabled
    if source and source.player and source.player.valid then
	
		-- Chaos Emeralds/Master Emerald shards - 1000 points
		if (target.type >= 118 and target.type <= 125) then
			earnPointsForObject(source.player, target, 1000)
			
		-- Monitor handling (Depends on the monitor)
        elseif target.info.flags & MF_MONITOR then
		
			-- Golden monitors - 1000 points
            if target.type >= 177 and target.type <= 189 then
				earnPointsForObject(source.player, target, 1000)
				return
			end
			
			-- Extra life monitors - 1000 points
			if target.type == 166 then
				earnPointsForObject(source.player, target, 1000)
				return
			end
			
            if target.type == 172 or target.type == 173 then return end -- No score monitors
            if target.type == 167 then return end -- No Eggman monitors
            earnPointsForObject(source.player, target, 500) -- All other monitors - 500 points
			
		-- Air bubbles - 500 points
		elseif target.type == MT_EXTRALARGEBUBBLE then
			local bubblesource = target
			for mobj in mobjs.iterate() do
				if mobj.type == MT_BUBBLES and target.x == mobj.x and target.y == mobj.y then
					bubblesource = mobj
				end
			end
			earnPointsForObject(source.player, target, 500)

		-- Everything else (rings, coins, spikes) - 10 points
		-- No emerald tokens, lost rings, or enemies/bosses!
		elseif not ((target.type == MT_TOKEN) or (target.type == MT_FLINGRING)
		or (target.flags & MF_ENEMY) or (target.flags & MF_BOSS)) then
			earnPointsForObject(source.player, target, 10)
		end
    end
end

addHook("MobjDamage", awardItemPoints)
addHook("MobjDeath", awardItemPoints)

-- Award 100 points when you damage (not kill) enemies/bosses
local function awardEnemyPoints(target, inflictor, source)
	if NoSBSystem() then return end -- Combo system is disabled
    if source and source.player and source.player.valid then
        if ((target.flags & MF_ENEMY) or (target.flags & MF_BOSS))
		and target.health and target.health > 1 then -- Ignoring when you finally kill them
			P_AddPlayerScore(source.player, 100)
		end
    end
end

addHook("MobjDamage", awardEnemyPoints)

-- Have the player earn the given points from the given stock type
local function earnPointsFromStock(p, stock, points)
	if p and p.valid and p[stock] then
		if points <= p[stock] then -- If there are ample points left in the stock
			P_AddPlayerScore(p, points) -- Award the points
			p[stock] = $ - points -- And remove from the stock
		else -- If there aren't enough points left in the stock
			P_AddPlayerScore(p, p[stock]) -- Award whatever points are left
			p[stock] = 0 -- Let stock points fall to zero
		end
	end
end

-- Award points for doing actions in-level
local function awardActionPoints(p)
	if NoSBSystem() then return end -- Combo system is disabled
    if not (p and p.valid) then return end
	
	-- Springs - 200 points (at max)
    if p.mo.eflags & MFE_SPRUNG then
		local sincesprung = p.realtime - p.sb_lastsprungtime
		if sincesprung > 3 * TICRATE then
			earnPointsFromStock(p, "sb_springscore", 200)
		else
			earnPointsFromStock(p, "sb_springscore", 50)
		end
		p.sb_lastsprungtime = p.realtime
	end

end

addHook("PlayerThink", awardActionPoints)

-- Award points for grinding
local function awardGrindPoints(p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid) then return end -- Valid player?

	-- If we're grinding
	if p.powers[pw_carry] == CR_GRINDRAIL then -- p.mo.GSgrinding or (p.mo.GSgrind and p.mo.GSgrind.grinding) then
		-- If we just started grinding
		if p.sb_grindlastframe == 0 then
			earnPointsFromStock(p, "sb_grindscore", 10)
			p.sb_grindstarttime = p.realtime
		-- If we're continuing to grind
		else
			if ((p.realtime - p.sb_grindstarttime) % TICRATE) == 0 then
				earnPointsFromStock(p, "sb_grindscore", 10)
			end
		end
	end
	
	-- Were we grinding on this frame?
	if p.powers[pw_carry] == CR_GRINDRAIL then -- p.mo.GSgrinding or (p.mo.GSgrind and p.mo.GSgrind.grinding) then
		p.sb_grindlastframe = 1
	else
		p.sb_grindlastframe = 0
	end
end

addHook("PlayerThink", awardGrindPoints)

-- Award continuous points for carry actions
local function awardCarryPoints(p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid) then return end -- Valid player?

	-- Carry actions to award points for
	local awardCarryTypes = {CR_ZOOMTUBE, CR_ROPEHANG, CR_MACESPIN,
							 CR_MINECART, CR_ROLLOUT, CR_DUSTDEVIL, CR_FAN}
	local validCarryAction = false
	for i = 1, #awardCarryTypes do
		if p.powers[pw_carry] == awardCarryTypes[i] then
			validCarryAction = true
		end
	end
	
	-- If the player is doing one of these carry actions
	if validCarryAction then
		-- If the player is just starting this carry type, award points
		if p.powers[pw_carry] != p.sb_lastcarrytype then
			p.sb_carrystarttime = p.realtime
			earnPointsFromStock(p, "sb_continuousscore", 10)
		-- If the player is continuing this carry type, award points every second
		elseif (p.realtime - p.sb_carrystarttime) % TICRATE == 0 then
			earnPointsFromStock(p, "sb_continuousscore", 10)
		end
	end

	-- Update the last carry type
	p.sb_lastcarrytype = p.powers[pw_carry]
end

addHook("PlayerThink", awardCarryPoints)

-- Award continuous points for sliding in a waterfall
local function awardSlidingPoints(p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid) then return end
	if p.pflags & PF_SLIDING then
		-- If the player is just starting to slide, award points
		if sb_slidelastframe then
			p.sb_slidestarttime = p.realtime
			earnPointsFromStock(p, "sb_continuousscore", 10)
		-- If the player is continuing to slide, award points every second
		elseif (p.realtime - p.sb_slidestarttime) % TICRATE == 0 then
			earnPointsFromStock(p, "sb_continuousscore", 10)
		end
	end
	p.sb_slidelastframe = p.pflags & PF_SLIDING
end

-- addHook("PlayerThink", awardSlidingPoints) -- THIS DOESN'T WORK, HAVE TO DO BELOW

-- Why the fuck do PF_SLIDING checks work in ThinkFrame but not PlayerThink????????
local function awardSlidingPointsMain()
	if NoSBSystem() then return end -- Combo system is disabled
	for player in players.iterate do
		awardSlidingPoints(player)
	end
end

addHook("ThinkFrame", awardSlidingPointsMain)

-- Are these two objects colliding on the z-axis?
local function zCollide(thing, tmthing)
	if tmthing.z == thing.z then
		return true
	elseif tmthing.z > thing.z then
		return (thing.z + thing.height) > tmthing.z
	elseif tmthing.z < thing.z then
		return (tmthing.z + tmthing.height) > thing.z
	end
end

-- Earn 10 points every second you push something
-- (This detection is inconsistent as hell because pushables fucking suck)
local function awardPushablePoints(thing, tmthing)
	if NoSBSystem() then return end -- Combo system is disabled
	if thing and thing.valid and tmthing and tmthing.valid
	and thing.flags & MF_PUSHABLE and tmthing.player
	and tmthing.player.powers[pw_pushing]
	and tmthing.player.realtime % TICRATE == 0
	and zCollide(thing, tmthing) then
		earnPointsFromStock(tmthing.player, "sb_continuousscore", 10)
	end
end

-- Search for all pushable objects
for i = 0, #mobjinfo - 1 do
    if mobjinfo[i] and (mobjinfo[i].flags & MF_PUSHABLE) then
		addHook("MobjCollide", awardPushablePoints, i)
    end
end

-- Award points for colliding with a tumbleweed/mine/pushable object
local function awardMobjCollidePoints(thing, tmthing)
	if NoSBSystem() then return end -- Combo system is disabled
	if thing and thing.valid and tmthing and tmthing.valid
	and tmthing.player and zCollide(thing, tmthing) then
		if thing.type == MT_NIGHTSBUMPER then -- 200 points for NiGHTS bumpers
			earnPointsForObject(tmthing.player, thing, 200)
		elseif thing.type == MT_BIGMINE then -- 100 points for mines
			earnPointsForObject(tmthing.player, thing, 100)
		else -- 10 points for tumbleweeds/other pushable objects
			earnPointsForObject(tmthing.player, thing, 10)
		end
	end
end

addHook("MobjCollide", awardMobjCollidePoints, MT_BIGTUMBLEWEED)
addHook("MobjCollide", awardMobjCollidePoints, MT_LITTLETUMBLEWEED)
addHook("MobjCollide", awardMobjCollidePoints, MT_BIGMINE)
addHook("MobjCollide", awardMobjCollidePoints, MT_NIGHTSBUMPER)

-- Award points for touching a star post/emblem (500/1000 points)
local function awardTouchSpecialPoints(special, toucher)
	if NoSBSystem() then return end -- Combo system is disabled
    if toucher and toucher.valid and special and special.valid
	and toucher.player and not toucher.player.bot
	and not toucher.player.sb_touchedobjects[special] then
		if special.type == MT_STARPOST then
			earnPointsForObject(toucher.player, special, 500)
			S_StartSound(toucher, sfx_strpst)
		end
		
		if special.type == MT_EMBLEM then
			earnPointsForObject(toucher.player, special, 1000)
			S_StartSound(toucher, sfx_ncitem)
		end
		
		-- JanaTools support
		if special.type == MT_DASHPANEL
		or special.type == MT_DASHRING
		or special.type == MT_DASHRAMP then
			earnPointsForObject(toucher.player, special, 200)
		end
    end
end

addHook("TouchSpecial", awardTouchSpecialPoints, MT_STARPOST)
addHook("TouchSpecial", awardTouchSpecialPoints, MT_EMBLEM)
addHook("TouchSpecial", awardTouchSpecialPoints, MT_DASHPANEL)
addHook("TouchSpecial", awardTouchSpecialPoints, MT_DASHRING)
addHook("TouchSpecial", awardTouchSpecialPoints, MT_DASHRAMP)

-- Spawn star posts in record attack
local function spawnStarPosts()
	if NoSBSystem() then return end -- Combo system is disabled
	if not modeattacking then return end
	for mapthing in mapthings.iterate do
		if mapthing.type == 502 then
			local spawnedpost = P_SpawnMobj(mapthing.x * FRACUNIT, mapthing.y * FRACUNIT, 0, MT_STARPOST)
			spawnedpost.z = spawnedpost.subsector.sector.floorheight + (mapthing.z * FRACUNIT)
			spawnedpost.angle = FixedAngle(mapthing.angle * FRACUNIT)
		end
	end
end

addHook("MapLoad", spawnStarPosts)

----- DISPLAY COMBO INFORMATION IN THE HUD -----

-- Animate the combo star based on how long the timer has left
local function animateComboStar(p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid and p.sb_combotimer) then return end
	if p.sb_combotimer > ((sb_combolength * 2) / 3) then
		if p.jointime % 1 == 0 then p.sb_comboframe = $ + 1 end
	elseif p.sb_combotimer > (sb_combolength / 3) then
		if p.jointime % 2 == 0 then p.sb_comboframe = $ + 1 end
	elseif p.sb_combotimer > 0 * TICRATE then
		if p.jointime % 3 == 0 then p.sb_comboframe = $ + 1 end
	end
	
	-- If we've looped over the frames, or if we're exiting, reset to the 1st frame
	if p.sb_comboframe > 7 or p.exiting then p.sb_comboframe = 1 end
end

addHook("PlayerThink", animateComboStar)

-- Displaying the combo star!
hud.add(function(v, p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid and p.sb_combotimer) then return end
	v.draw(320 - 24 - 5, 0 + 5, v.cachePatch("ECS" .. tostring(p.sb_comboframe)), V_SNAPTOTOP | V_SNAPTORIGHT)
end)

-- Helper function for drawing recordscore
local function drawRecordScore(v, p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid) then return end
	
	local rs_draw = tostring(p.recordscore)
	rs_draw = string.reverse(rs_draw)
	local recordlen = string.len(rs_draw)
	
	for i = 1, recordlen, 1 do
		v.drawScaled((320 - 37 - (11 * (i - 1))) * FRACUNIT, 20 * FRACUNIT, (FRACUNIT * 3) / 4,
					 v.cachePatch("SA2TL" .. tostring(string.sub(rs_draw, i, i))),
				     V_SNAPTOTOP | V_SNAPTORIGHT)
    end
end

-- Displaying the earned combo!
hud.add(function(v, p)
	if NoSBSystem() then return end -- Combo system is disabled
	if not (p and p.valid and p.sb_lastcomboend) then return end
	
	-- Initialize some necessary variables
	local transflaglist = {V_10TRANS, V_20TRANS, V_30TRANS,
						   V_40TRANS, V_50TRANS, V_60TRANS,
						   V_70TRANS, V_80TRANS, V_90TRANS}
	local sincecomboend = p.jointime - p.sb_lastcomboend
	local earnflags = V_SNAPTOTOP | V_SNAPTORIGHT | V_ALLOWLOWERCASE
	-- v.drawString(320 - 32, 16, p.recordscore, earnflags, "right")
	drawRecordScore(v, p)
	
	-- v.drawScaled(fixed_t x, fixed_t y, fixed_t scale, patch_t patch, [int flags, [colormap c]])
	
	local fadestart = (3 * TICRATE) / 2
	
	-- Determine display flags based on timer
	if sincecomboend < fadestart + 10 then
		local transindex, currenttrans
		if sincecomboend > fadestart then 
			transindex = sincecomboend - fadestart
			currenttrans = transflaglist[transindex]
			earnflags = $ | currenttrans
		end
		
		/*
		-- Render more specific multiplier information
		v.drawString(320 - 8, 34, p.sb_lastcomboscore, earnflags, "right")	
		local wholeNumber = p.sb_lastcombomulti / 2
		local decimalNumber = 0
		if p.sb_lastcombomulti % 2 == 1 then decimalNumber = 5 end
		local multiplierValue = tostring(wholeNumber) .. "." .. tostring(decimalNumber)
		v.drawString(320 - 8, 45, "x" .. multiplierValue, earnflags, "right")
		*/
		
		-- Draw information signifying that you got bonus points
		if p.sb_lastcombomulti > 0 then
			local labelindex = p.sb_lastcombomulti
			if labelindex > 10 then labelindex = 10 end
			local combolabel = v.cachePatch("CSCL" .. tostring(labelindex))
			
			local labelxcoord = 160 - (combolabel.width / 4)
			v.drawScaled(labelxcoord * FRACUNIT, 8 * FRACUNIT, FRACUNIT / 2, combolabel, currenttrans)
			-- v.drawString(160, 20, (p.sb_lastcomboscore * p.sb_lastcombomulti) / 2, currenttrans, "center")
			v.drawString(160, 20, p.sb_lastcomboscore, currenttrans, "center")
		end
	end
end)