-- Script for defining Metal Harbor stuff

freeslot("MT_MHLIGHT", "S_MHLIGHT", "SPR_MHLG")

states[S_MHLIGHT] = {
	sprite = SPR_MHLG,
    frame = A | FF_PAPERSPRITE,
	tics = -1,
	nextstate = S_MHLIGHT
}

mobjinfo[MT_MHLIGHT] = {
	--$Name Metal Harbor Light
	--$Category Metal Harbor
	--$Sprite MHLGA0
	doomednum = 6351,
    spawnstate = S_MHLIGHT,
    radius = 16 * FRACUNIT,
    height = 128 * FRACUNIT,
    flags = MF_SCENERY | MF_NOCLIP | MF_NOGRAVITY
}

freeslot("MT_MHRCKT", "S_MHRCKT", "SPR_MRKT")

states[S_MHRCKT] = {
	sprite = SPR_MRKT,
    frame = A,
	tics = -1,
	nextstate = S_MHRCKT
}

mobjinfo[MT_MHRCKT] = {
	--$Name Metal Harbor Rocket
	--$Category Metal Harbor
	--$Sprite MRKTA1
	doomednum = 6352,
    spawnstate = S_MHRCKT,
    radius = 96 * FRACUNIT,
    height = 1024 * FRACUNIT,
    flags = MF_SCENERY | MF_NOCLIP | MF_NOGRAVITY
}

freeslot("MT_MHBRRL", "S_MHBRRL", "SPR_MBRL")

states[S_MHBRRL] = {
	sprite = SPR_MBRL,
    frame = A,
	tics = -1,
	nextstate = S_MHBRRL
}

mobjinfo[MT_MHBRRL] = {
	--$Name Metal Harbor Barrel
	--$Category Metal Harbor
	--$Sprite MBRLA1A5
	doomednum = 6353,
    spawnstate = S_MHBRRL,
    radius = 64 * FRACUNIT,
    height = 320 * FRACUNIT,
    flags = MF_SCENERY | MF_NOCLIP | MF_NOGRAVITY
}

-- Bottom extender for Metal Harbor barrels

freeslot("MT_MHBRRLBT", "S_MHBRRLBT", "SPR_MBRB")

states[S_MHBRRLBT] = {
	sprite = SPR_MBRB,
    frame = A,
	tics = -1,
	nextstate = S_MHBRRLBT
}

mobjinfo[MT_MHBRRLBT] = {
	--$Name Metal Harbor Barrel Pole
	--$Category Metal Harbor
	--$Sprite MBRBA1A5
	doomednum = 6354,
    spawnstate = S_MHBRRLBT,
    radius = 64 * FRACUNIT,
    height = 32 * FRACUNIT,
    flags = MF_SCENERY | MF_NOCLIP | MF_NOGRAVITY
}

-- Fill the bottom of the barrel with pole
local function barrelBottomFill(mo)
	local bottomz = mo.z - FixedMul(mo.height, mo.spriteyscale)
	local floorz = mo.subsector.sector.floorheight
	local poleheight = FixedMul(32 * FRACUNIT, mo.spriteyscale)
	local spawnheight = bottomz
	
	while spawnheight > floorz do
		local pole = P_SpawnMobj(mo.x, mo.y, spawnheight, MT_MHBRRLBT)
		pole.spritexscale = mo.spritexscale
		pole.spriteyscale = mo.spriteyscale
		pole.angle = mo.angle
        spawnheight = $ - poleheight
    end
end

addHook("MapThingSpawn", barrelBottomFill, MT_MHBRRL)

-- Trigger for starting the rocket climb
local function rocketLaunchTimer(line, mo)
	if mo and mo.player and not mo.player.mhtimer then
		mo.player.mhtimer = mo.player.realtime
		S_StartSound(mo, sfx_alarm, mo.player)
	end
end

addHook("LinedefExecute", rocketLaunchTimer, "MHTIMER")

-- How much time do we have for the rocket launch?
local rocketlaunchtimer = 20 * TICRATE

-- Thinker for the rocket launch alarm
local function rocketLaunchAlarm(p)
	if p and p.valid and p.mhtimer
	and mapheaderinfo[gamemap].lvlttl == "Metal Harbor"
	and p.playerstate == PST_LIVE then
		local timertime = p.realtime - p.mhtimer
		if timertime % TICRATE == 0 then
			S_StartSound(p.mo, sfx_alarm, p)
		end
		
		local timeleft = p.mhtimer + rocketlaunchtimer - p.realtime
		if timeleft == 0 then
			P_FlashPal(p, 4, 2 * TICRATE)
			S_StartSound(p.mo, sfx_bebomb)
			P_KillMobj(p.mo)
		end
	end
end

addHook("PlayerThink", rocketLaunchAlarm)

-- Function for displaying the rocket timer
local function rocketTimerDisplay(v, p)
	if p and p.valid and p.mhtimer
	and mapheaderinfo[gamemap].lvlttl == "Metal Harbor" then
		local timeleft = p.mhtimer + rocketlaunchtimer - p.realtime
		if timeleft < 0 then timeleft = 0 end
		local secsleft = G_TicsToSeconds(timeleft)
		local csecsleft = G_TicsToCentiseconds(timeleft)
		if string.len(csecsleft) == 1 then
			csecsleft = tostring(csecsleft) .. '0'
		end
		local timestring = tostring(secsleft) .. '.' .. tostring(csecsleft)
		v.drawString(160, 180, timestring, V_PERPLAYER | V_ALLOWLOWERCASE | V_REDMAP, "center")
	end
end

hud.add(rocketTimerDisplay)

-- Trigger for finishing the rocket climb
local function rocketLaunchTrigger(line, mo)
	if mo and mo.player and not mo.player.mhrocket
	and mo.player.powers[pw_carry] == CR_ROPEHANG then
		mo.player.mhrocket = mo.player.realtime
		mo.player.mhtimer = nil
		S_StartSound(mo, sfx_rumble, mo.player)
	end
end

addHook("LinedefExecute", rocketLaunchTrigger, "MHROCKET")

-- Thinker for the rocket launch
local function rocketLaunchThinker(p)
	if p and p.valid and p.mhrocket
	and mapheaderinfo[gamemap].lvlttl == "Metal Harbor" then
		local rockettime = p.realtime - p.mhrocket
		-- Charging the rocket
		if rockettime <= 3 * TICRATE then
			if rockettime == 0 then
				S_StartSound(p.mo, sfx_rumble, p)
				-- WHY DOESN'T THIS EARTHQUAKE WORK???????????????????????????????????????????????
				P_StartQuake(640 * FRACUNIT, 3 * TICRATE, {p.mo.x, p.mo.y, p.mo.z}, 64 * FRACUNIT)
			end
			p.pflags = $ | PF_FULLSTASIS
		-- Liftoff!
		elseif rockettime == (3 * TICRATE) + 1 then
			p.powers[pw_carry] = CR_NONE
			p.mo.momy = 52 * FRACUNIT
			p.mo.momz = 60 * FRACUNIT - (FRACUNIT / 2)
			local newy = p.mo.y + (p.mo.momy * 2)
			local newz = p.mo.z + (p.mo.momz * 2)
			P_MoveOrigin(p.mo, p.mo.x, newy, newz)
			P_FlashPal(p, 4, TICRATE / 3)
			S_StartSound(p.mo, sfx_bebomb)
			S_StopSoundByID(p.mo, sfx_rumble)
			S_StopMusic(p)
			p.mo.state = S_PLAY_FALL
			p.pflags = $ | PF_FULLSTASIS
		-- Lock controls until we reach the end
		elseif rockettime <= (13 * TICRATE) / 2 then
			S_StopMusic(p) -- Don't let Metal Harbor come back!
			p.pflags = $ | PF_FULLSTASIS
		elseif rockettime <= (34 * TICRATE) / 4 then
			p.pflags = $ | PF_FULLSTASIS
		end
	end
end

addHook("PlayerThink", rocketLaunchThinker)

-- Reset rocket launch on player spawn
local function rocketLaunchReset(p)
	if p and p.valid then
		p.mhrocket = nil
		p.mhtimer = nil
	end
end

addHook("PlayerSpawn", rocketLaunchReset)