-- Script for allchars support (SP-side)

-- Delete objects with ALLCHARS flag if allchars is not enabled
-- If allchars is enabled, makes the objects transparent
local function allCharsLoad()
	for mo in mobjs.iterate() do
		if mo and mo.valid and mo.spawnpoint
		and (mo.spawnpoint.stringargs[0] == "ALLCHARS"
		or mo.spawnpoint.stringargs[1] == "ALLCHARS") then
			if not CV_FindVar("allchars")
			or CV_FindVar("allchars").value == 0 then
				P_RemoveMobj(mo)
			end
		end
	end
end

addHook("MapLoad", allCharsLoad)

-- Is this an allchars object?
local function allCharsObject(mo)
	if mo and mo.valid and mo.spawnpoint
	and (mo.spawnpoint.stringargs[0] == "ALLCHARS"
	or mo.spawnpoint.stringargs[1] == "ALLCHARS") then
		return true
	else
		return false
	end
end

-- Make allchars objects flash like a hologram
local function allCharsHologram(mo)
	if mo and mo.valid then
		if leveltime % 2 < 1 then
			mo.frame = $ & ~FF_TRANS20
			mo.frame = $ | FF_TRANS40
		else
			mo.frame = $ & ~FF_TRANS40
			mo.frame = $ | FF_TRANS20
		end
	end
end

-- Make allchars springs transparent
local function transSprings(mo)
	if allCharsObject(mo) then
		allCharsHologram(mo)
	end
end

addHook("MobjThinker", transSprings, MT_YELLOWSPRING)
addHook("MobjThinker", transSprings, MT_REDSPRING)
addHook("MobjThinker", transSprings, MT_BLUESPRING)
addHook("MobjThinker", transSprings, MT_YELLOWDIAG)
addHook("MobjThinker", transSprings, MT_REDDIAG)
addHook("MobjThinker", transSprings, MT_BLUEDIAG)
addHook("MobjThinker", transSprings, MT_YELLOWHORIZ)
addHook("MobjThinker", transSprings, MT_REDHORIZ)
addHook("MobjThinker", transSprings, MT_BLUEHORIZ)
addHook("MobjThinker", transSprings, MT_YELLOWBOOSTER)
addHook("MobjThinker", transSprings, MT_REDBOOSTER)

local mainbooster = nil

-- Save the main booster object to its parts as a field
local function getMainBooster(mo)
	if mo.type == MT_YELLOWBOOSTER or mo.type == MT_REDBOOSTER then
		mainbooster = mo
	elseif mo.type == MT_BOOSTERSEG or mo.type == MT_BOOSTERROLLER then
		mo.booster = mainbooster
		if not mo.booster.pieces then
			mo.booster.pieces = {mo}
		else
			table.insert(mo.booster.pieces, mo)
		end
	end
end

addHook("MobjSpawn", getMainBooster, MT_YELLOWBOOSTER)
addHook("MobjSpawn", getMainBooster, MT_REDBOOSTER)
addHook("MobjSpawn", getMainBooster, MT_BOOSTERSEG)
addHook("MobjSpawn", getMainBooster, MT_BOOSTERROLLER)

-- Make allchars booster parts transparent
local function transBoosters(mo)
	if mo and mo.valid and mo.booster
	and allCharsObject(mo.booster) then
		if CV_FindVar("allchars")
		and CV_FindVar("allchars").value == 1 then
			allCharsHologram(mo)
		else
			P_RemoveMobj(mo)
		end
	end
end

addHook("MobjThinker", transBoosters, MT_BOOSTERSEG)
addHook("MobjThinker", transBoosters, MT_BOOSTERROLLER)

-- Make allchars cubes transparent
local function transCubes(mo)
	if allCharsObject(mo) then
		allCharsHologram(mo)
		if mo.cubefaces then
			for i = 1, 6 do
				if mo.cubefaces[i] and mo.cubefaces[i].frame then
					allCharsHologram(mo.cubefaces[i])
				end
			end
		end
	end
end

addHook("MobjThinker", transCubes, MT_SA2STEELBOX)

-- Make allchars cubes transparent
local function transFaces(mo)
	if mo and mo.valid and mo.target
	and allCharsObject(mo.target) then
		allCharsHologram(mo)
	end
end

addHook("MobjThinker", transFaces, MT_SA2STEELFACE)

freeslot("MT_PULLEYBASE")

-- Makes allchars pulleys transparent
local function transPulleys(mo)
	if allCharsObject(mo) then
		allCharsHologram(mo)
		if mo.rope and mo.rope.frame then allCharsHologram(mo.rope) end
		if mo.handle and mo.handle.frame then allCharsHologram(mo.handle) end
	end
end

addHook("MobjThinker", transPulleys, MT_PULLEYBASE)

local mpWarningMessage = false

-- Warning message about the multiplayer patch
local function mpWarning(p)
	if (not CV_FindVar("allchars") or CV_FindVar("allchars").value == 0)
	and p and p.mo and p.mo.skin != "sonic" and p.mo.skin != "tails"
	and p.mo.skin != "knuckles" and p.mo.skin != "shadow"
	and p.mo.skin != "eggman" and p.mo.skin != "rouge"
	and not mpWarningMessage then
		CONS_Printf(p, "\x85" .. "WARNING: " .. "\x80" .. "If you want to play as other characters, you must download and add the multiplayer patch! If you already have the multiplayer patch, simply add it in the addons menu, or run the \"multiplayer\" command.")
		mpWarningMessage = true
	end
end

addHook("PlayerThink", mpWarning)

-- Re-print the warning message every time the player spawns
local function mpWarningReset(p)
	mpWarningMessage = false
end

addHook("PlayerSpawn", mpWarningReset)

-- Shortcut function to add the multiplayer patch
local function sa2BlastMP(p)
	COM_BufInsertText(p, "addfile SCL_SA2-Blast_MP-Patch_v1.2.pk3") 
end

COM_AddCommand("multiplayer", sa2BlastMP)