-- Script for emerald hunt stuff

freeslot("MT_SA2EMSPAWN", "S_SA2EMSPAWN", "SPR_SBSH")

states[S_SA2EMSPAWN] = {
	sprite = SPR_NULL,
    frame = A,
	tics = -1,
	nextstate = S_SA2EMSPAWN
}

mobjinfo[MT_SA2EMSPAWN] = {
	--$Name SA2 Emerald Hunt Spawn
	--$Category SA2:Blast Objects
	--$Sprite SBSHA0
	doomednum = 6400,
    spawnstate = S_SA2EMSPAWN,
    radius = 8 * FRACUNIT,
    height = 44 * FRACUNIT,
    flags = MF_NOGRAVITY
}

freeslot("MT_SA2EMSHARD", "S_SA2EMSHARDA", "S_SA2EMSHARDB", "S_SA2EMSHARDC")

states[S_SA2EMSHARDA] = {
	sprite = SPR_SBSH,
    frame = A,
	tics = -1,
	nextstate = S_SA2EMSHARDA
}

states[S_SA2EMSHARDB] = {
	sprite = SPR_SBSH,
    frame = B,
	tics = -1,
	nextstate = S_SA2EMSHARDB
}

states[S_SA2EMSHARDC] = {
	sprite = SPR_SBSH,
    frame = C,
	tics = -1,
	nextstate = S_SA2EMSHARDC
}

mobjinfo[MT_SA2EMSHARD] = {
    spawnstate = S_SA2EMSHARDA,
    radius = 8 * FRACUNIT,
    height = 44 * FRACUNIT,
    flags = MF_SPECIAL | MF_NOGRAVITY
}

-- Reset collected emeralds upon player spawn
local function collectedEmeralds(p)
	if not (p and p.valid) then return end
	p.sa2gotemeralds = {}
	p.sa2emeraldtimer = p.realtime
end

addHook("PlayerSpawn", collectedEmeralds)

-- Handle what happens when you touch a shard
local function touchShard(special, toucher)
	if toucher.player.bot then return end -- No bots!
	if not (toucher.player and toucher.player.valid -- Valid player?
	and toucher.player.playerstate == PST_LIVE) then return end
	
	-- If the shard hasn't been collected and the player isn't finished
	if not toucher.player.sa2gotemeralds[special.shardid]
	and not (toucher.player.pflags & PF_FINISHED) then
		P_SpawnMobjFromMobj(special, 0, 0, 2 * FRACUNIT, MT_SPARK) -- Spawn a sparkle object
		S_StartSound(special, sfx_cgot) -- Play the collected sound
		
		-- Score based on how fast you found the emerald
		local p = toucher.player
		if p.realtime - p.sa2emeraldtimer <= 30 * TICRATE then
			P_AddPlayerScore(toucher.player, 10000) -- 55000 bonus
		elseif p.realtime - p.sa2emeraldtimer <= 40 * TICRATE then
			P_AddPlayerScore(toucher.player, 8000) -- 36000 bonus
		elseif p.realtime - p.sa2emeraldtimer <= 50 * TICRATE then
			P_AddPlayerScore(toucher.player, 6000) -- 21000 bonus
		elseif p.realtime - p.sa2emeraldtimer <= 60 * TICRATE then
			P_AddPlayerScore(toucher.player, 4000) -- 10000 bonus
		else
			P_AddPlayerScore(toucher.player, 2000) -- 3000 bonus
		end
		p.sa2emeraldtimer = p.realtime

		toucher.player.sa2gotemeralds[special.shardid] = true -- And mark shard as collected
		
		-- Finish the stage if we've got all the emeralds
		local finishstage = true
		for i = 1, tonumber(mapheaderinfo[gamemap].sa2ecount) do
			finishstage = finishstage and p.sa2gotemeralds[i]
		end
		if finishstage then P_DoPlayerFinish(toucher.player) end
		
	end
	return true
end

addHook("TouchSpecial", touchShard, MT_SA2EMSHARD)

-- Make collected shards display as transparent
local function transShards(mo)
	if not (mo and mo.valid) then return end
	
	-- If the shard has been collected or the player is finished
	if displayplayer.sa2gotemeralds[mo.shardid]
	or displayplayer.pflags & PF_FINISHED then
		mo.flags2 = $ | MF2_DONTDRAW -- Make it disappear!
	else
		mo.flags2 = $ & ~MF2_DONTDRAW -- Otherwise, make it reappear
	end
end

addHook("MobjThinker", transShards, MT_SA2EMSHARD)

-- Get the distance between two 3D points
local function calculate_distance(coord1, coord2)
	local xdiff = ((coord1[1] / FRACUNIT) - (coord2[1] / FRACUNIT))^2
	local ydiff = ((coord1[2] / FRACUNIT) - (coord2[2] / FRACUNIT))^2
	local zdiff = ((coord1[3] / FRACUNIT) - (coord2[3] / FRACUNIT))^2
	local result = FixedSqrt(xdiff + ydiff + zdiff)
    return result
end

-- Sort descending by the 4th element in the coordinate list
local function descend_sort_by_4th(element_a, element_b)
	return element_a[4] > element_b[4]
end

-- Choose emerald shard spawn locations based on:
-- chosen_emeralds -- The already chosen emerald coordinates (Initialize as an empty list)
-- emerald_coords -- The possible remaining emerald coordinates (Initialize as a list of 3-element lists)
-- emerald_count -- The desired number of emeralds to choose (Initialize as whatever number you want lol)
local function choose_emeralds(chosen_emeralds, emerald_coords, emerald_count)
	
	if #emerald_coords < (emerald_count - #chosen_emeralds) then
		print("ERROR: Not enough possible emerald locations!")
		return
	end
	
    -- If the number of chosen emeralds match, then we're done!
    if #chosen_emeralds == emerald_count then
        return chosen_emeralds
    end
    
    -- If this is the first emerald we're choosing, just choose it randomly
    if #chosen_emeralds == 0 then
        local chosen_coordinate_index = P_RandomRange(1, #emerald_coords)
        local chosen_coordinate = emerald_coords[chosen_coordinate_index]
        table.insert(chosen_emeralds, chosen_coordinate)
		table.remove(emerald_coords, chosen_coordinate_index)
        return choose_emeralds(chosen_emeralds, emerald_coords, emerald_count)
    end
	
	-- If we've already chosen some emeralds, then...
	if #chosen_emeralds > 0 then
	
		-- Calculate the distances betweeen the already chosen emeralds
		-- And the emeralds that are candidates to be chosen
		for i = 1, #emerald_coords do
			for j = 1, #chosen_emeralds do
				local emerald_distance = calculate_distance(emerald_coords[i], chosen_emeralds[j])
				if j == 1 then
					emerald_coords[i][4] = emerald_distance
				else
					if emerald_distance < emerald_coords[i][4] then
						emerald_coords[i][4] = emerald_distance
					end
				end
			end
			/*
			for j = 1, #chosen_emeralds do
				local emerald_distance = calculate_distance(emerald_coords[i], chosen_emeralds[j])
				emerald_coords[i][4] = $ + emerald_distance
			end
			*/
		end
	
		-- Choose randomly from the Xth of the emeralds that are farther away
		table.sort(emerald_coords, descend_sort_by_4th)
		-- local chosen_coordinate_index = P_RandomRange(1, (#emerald_coords / 2) + 1)
		local chosen_coordinate_index = P_RandomRange(1, (#emerald_coords / (#chosen_emeralds + 1)) + 1)
		
		local chosen_coordinate = emerald_coords[chosen_coordinate_index]
        table.insert(chosen_emeralds, chosen_coordinate)
		table.remove(emerald_coords, chosen_coordinate_index)
        return choose_emeralds(chosen_emeralds, emerald_coords, emerald_count)
	end
end

-- Sort ascending by the emerald shard ID
local function ascend_emerald_id(element_a, element_b)
	return element_a.shardid < element_b.shardid
end

-- Load the emeralds into player data upon player spawn
local function loadEmeralds(p)
	
	-- Only if defined as an emerald hunt map in the level header
	if not (p and p.valid) then return end
	if not mapheaderinfo[gamemap].sa2ecount then
		p.sa2emeraldtable = nil
		return
	end
	
	-- Find all the spawned emeralds
	local emeraldtable = {}
	for mobj in mobjs.iterate() do
		if mobj.type == MT_SA2EMSHARD then
			table.insert(emeraldtable, mobj)
		end
	end
	
	-- Sort the emeralds by ID
	table.sort(emeraldtable, ascend_emerald_id)
	
	-- Assign emerald table to player data
	p.sa2emeraldtable = emeraldtable
	
end

addHook("PlayerSpawn", loadEmeralds)

-- Spawn the emeralds upon map load
local function spawnEmeralds()

	-- Only if defined as an emerald hunt map in the level header
	if not mapheaderinfo[gamemap].sa2ecount then return end

	-- Initialize some variables
	local emeraldmobjs = {}
	local emeraldcoords = {}
	
	-- Gather the coordinates for each emerald spawn
	for mobj in mobjs.iterate() do
		if mobj.type == MT_SA2EMSPAWN then
			-- Remove floating emeralds if allchars is true
			if not (CV_FindVar("allchars") and CV_FindVar("allchars").value == 1
			and mobj.z > mobj.floorz + (64 * FRACUNIT)) then
				table.insert(emeraldmobjs, mobj)
				table.insert(emeraldcoords, {mobj.x, mobj.y, mobj.z})
			else
				P_RemoveMobj(mobj)
			end
		end
	end
	
	-- Throw it into the function for choosing coords
	local chosencoords = choose_emeralds({}, emeraldcoords, tonumber(mapheaderinfo[gamemap].sa2ecount))
	
	-- Iterate over the chosen coordinates, then iterate over the emerald
	-- objects to find one with matching coordinates and spawn a shard there
	-- (MY ALGO PROFESSOR WILL FUCKING KILL ME BUT THIS IS HOW IT WORKS)
	for i = 1, #chosencoords do
		for j = 1, #emeraldmobjs do
			if emeraldmobjs[j].x == chosencoords[i][1]
			and emeraldmobjs[j].y == chosencoords[i][2]
			and emeraldmobjs[j].z == chosencoords[i][3] then
				local shard = P_SpawnMobjFromMobj(emeraldmobjs[j], 0, 0, 0, MT_SA2EMSHARD)
				if i % 3 == 0 then shard.state = S_SA2EMSHARDA end
				if i % 3 == 1 then shard.state = S_SA2EMSHARDB end
				if i % 3 == 2 then shard.state = S_SA2EMSHARDC end
				shard.shardid = i
			end
			-- if i == 1 then print(tostring(emeraldmobjs[j].x) .. " " .. tostring(emeraldmobjs[j].y) .. " " .. tostring(emeraldmobjs[j].z)) end
		end
	end
	
	-- Load emeralds into each player data
	for player in players.iterate()
		loadEmeralds(player)
	end

end

addHook("MapLoad", spawnEmeralds)

-- Draw emerald hunt UI
local function drawEmeraldHunt(v, p)
	
	-- Only if defined as an emerald hunt map in the level header
	if not mapheaderinfo[gamemap].sa2ecount then return end
	local emeraldcount = tonumber(mapheaderinfo[gamemap].sa2ecount)
	if p.pflags & PF_FINISHED then return end

	-- Cache tracking sprites
	local homing1 = v.cachePatch("HOMING1")
	local homing2 = v.cachePatch("HOMING2")
	local homing3 = v.cachePatch("HOMING3")
	local homing4 = v.cachePatch("HOMING4")
	local homing5 = v.cachePatch("HOMING5")
	local homing6 = v.cachePatch("HOMING6")
	
	-- Draw the tracking sprites
	local basecoord = 152 - ((emeraldcount - 1) * 10)
	local tracknoise = nil
	for i = 1, emeraldcount do
		local shard = p.sa2emeraldtable[i]
		
		-- Only if the shard hasn't been collected
		if not p.sa2gotemeralds[shard.shardid] then
			local distance = calculate_distance({p.mo.x, p.mo.y, p.mo.z}, 
												{shard.x, shard.y, shard.z}) / 256
												
			-- Determine sprite based on distance
			local tracksprite = nil
			if distance >= 3072 then
				tracksprite = homing1
			elseif distance >= 2048 then
				tracksprite = homing2
				if not tracknoise or tracknoise > 35 then tracknoise = 35 end
			elseif distance >= 1024 then
				tracksprite = homing3
				if not tracknoise or tracknoise > 30 then tracknoise = 30 end
			elseif distance >= 512 then
				tracksprite = homing4
				if not tracknoise or tracknoise > 20 then tracknoise = 20 end
			elseif distance >= 128 then
				tracksprite = homing5
				if not tracknoise or tracknoise > 10 then tracknoise = 10 end
			elseif distance >= 0 then
				tracksprite = homing6
				if not tracknoise or tracknoise > 5 then tracknoise = 5 end
			end
			
			-- Draw it!
			v.draw(basecoord + (20 * (i - 1)), 168, tracksprite, V_PERPLAYER)
		end
	end
	
	-- Make the beeping noise based on the closest shard
	if tracknoise and p.realtime % tracknoise == 0
	and not paused and not menuactive and not stoppedclock then
		S_StartSound(p.mo, sfx_emfind, p)
	end

end

hud.add(drawEmeraldHunt)