-- Miscellaneous SA2:Blast stuff

freeslot("MT_GUNHUNTER")

-- Make GUN Hunters unable to see you from far away
local function gunHunterDistance(mo)
	if mo and mo.valid and mo.GUN_HunterTable and mo.target then
		local xydist = P_AproxDistance(mo.x - mo.target.x, mo.y - mo.target.y)
		local xyzdist = P_AproxDistance(xydist, mo.z - mo.target.z)
		if (not (mo.skin == "tails" or mo.skin == "eggman") and xyzdist > 768 * FRACUNIT) or
		   ((mo.skin == "tails" or mo.skin == "eggman") and xyzdist > 1024 * FRACUNIT) then
			mo.target = nil
			mo.GUN_HunterTable.awakenanim = 0
			mo.GUN_HunterTable.aimanim = 0
			mo.GUN_HunterTable.recoilanim = 0
			-- mo.GUN_HunterTable.landanim = 0
		end
	end
end

addHook("MobjThinker", gunHunterDistance, MT_GUNHUNTER)

-- Recolor an object
local function recolorObjects()
    for mapthing in mapthings.iterate do
		if mapthing.mobj and mapthing.stringargs[0] == "RECOLOR" then
			mapthing.mobj.colorized = true
			mapthing.mobj.color = R_GetColorByName(mapthing.stringargs[1])		
		end
    end
end

addHook("MapLoad", recolorObjects)

-- Set momentum upon speed pad touch
local function sa2bSpeedPad(line, mo, sector)
	if mo and mo.valid and mo.player then
		mo.momx = line.args[0] * FRACUNIT
		mo.momy = line.args[1] * FRACUNIT
		mo.momz = line.args[2] * FRACUNIT
		mo.state = S_PLAY_FALL
		S_StartSound(mo, sfx_spdpad)
		mo.player.pflags = $ | PF_STASIS
		if mo.player.pflags & PF_SPINNING then
			mo.player.pflags = $ & ~PF_SPINNING
		end
		mo.player.sa2speedpad = leveltime
		
		-- Compatibility with SBCombo
		local p = mo.player
		if p.sb_touchedobjects
		and not p.sb_touchedobjects[line] then
			P_AddPlayerScore(p, 200)
			p.sb_touchedobjects[line] = true
		end
	end
end

addHook("LinedefExecute", sa2bSpeedPad, "SPEEDPADZ")

-- Reset speed pad touch time
local function sa2bSpeedPadReset(p)
	if p and p.valid then p.sa2speedpad = nil end
end

addHook("PlayerSpawn", sa2bSpeedPadReset)

-- Camera modification when going through loops/zoom tubes
local function sa2LoopCamera(p)
	if not (p and p.valid) then return end
	if p.sa2lastzoom == nil then p.sa2lastzoom = false end
	if p.bot then return end
	
	-- Starting a zoom tube
	if p.powers[pw_carry] == CR_ZOOMTUBE and not p.sa2lastzoom then
		local cam_speed = CV_FindVar("cam_speed")
		local cam_dist = CV_FindVar("cam_dist")
		local cam_height = CV_FindVar("cam_height")
		
		p.sa2camvalue = cam_speed.string
		p.sa2camdist = cam_dist.string
		p.sa2camheight = cam_height.string
		
		COM_BufInsertText(p, "cam_speed 0.05")
		COM_BufInsertText(p, "cam_dist 640")
		COM_BufInsertText(p, "cam_height -128")
	-- Reset camera once zoom tube is done
	elseif p.powers[pw_carry] != CR_ZOOMTUBE and p.sa2lastzoom then
		COM_BufInsertText(p, "cam_speed " .. tostring(p.sa2camvalue))
		COM_BufInsertText(p, "cam_dist " .. tostring(p.sa2camdist))
		COM_BufInsertText(p, "cam_height " .. tostring(p.sa2camheight))
		p.pflags = $ & ~PF_SPINNING
		-- p.pflags = $ | PF_JUMPED
		-- p.pflags = $ | PF_THOKKED
		p.mo.state = S_PLAY_SPRING
	end
	
	p.sa2lastzoom = p.powers[pw_carry] == CR_ZOOMTUBE
end

addHook("PlayerThink", sa2LoopCamera)

-- Reset camera values on player spawn
local function sa2ResetCamera(p)
	if p and p.valid
	and p.powers[pw_carry] != CR_ZOOMTUBE then
		local cam_speed = CV_FindVar("cam_speed")
		local cam_dist = CV_FindVar("cam_dist")
		local cam_height = CV_FindVar("cam_height")
		
		if cam_speed.string == "0.05" then
			COM_BufInsertText(p, "cam_speed 0.3")
		end
		
		if cam_dist.string == "640"
		and cam_height.string == "-128" then
			COM_BufInsertText(p, "cam_dist 192")
			COM_BufInsertText(p, "cam_height 40")
		end
	end
end

addHook("PlayerSpawn", sa2ResetCamera)

-- Option for objects to ignore gravity (WHY IS THIS NOT IN VANILLA)
local function noGravityObjects(mo, mapthing)
	if mapthing.stringargs[0] == "NOGRAVITY" then
		mo.flags = $ | MF_NOGRAVITY
	end
end

addHook("MapThingSpawn", noGravityObjects)

-- Makes monitors float by default
local function floatMonitors(mo)
	if not (mo and mo.valid) then return end
    mo.flags = $ | MF_NOGRAVITY
end

for i = 0, #mobjinfo - 1
    if mobjinfo[i] and (mobjinfo[i].flags & MF_MONITOR) then
        addHook("MapThingSpawn", floatMonitors, i)
    end
end

-- Are these two objects colliding on the z-axis?
local function zCollide(thing, tmthing)
	if tmthing.z == thing.z then
		return true
	elseif tmthing.z > thing.z then
		return (thing.z + thing.height) > tmthing.z
	elseif tmthing.z < thing.z then
		return (tmthing.z + tmthing.height) > thing.z
	end
end

-- Force-spin on boosters (Because Styles breaks this for some reason)
local function boosterSpin(thing, tmthing)
	if thing and thing.valid
	and tmthing and tmthing.valid
	and tmthing.player and zCollide(thing, tmthing)
	and thing.spawnpoint and thing.spawnpoint.args[0] then
		tmthing.player.sa2boosterspin = true
	end
end

addHook("MobjCollide", boosterSpin, MT_YELLOWBOOSTER)
addHook("MobjCollide", boosterSpin, MT_REDBOOSTER)

-- Force-spin on boosters (player-side)
local function boosterSpinPlayer(p)
	if p and p.valid and p.sa2boosterspin then
		p.pflags = $ | PF_SPINNING
		p.mo.state = S_PLAY_ROLL
		p.sa2boosterspin = false
	end
end

addHook("PlayerThink", boosterSpinPlayer)

freeslot("sfx_extlif")
sfxinfo[sfx_extlif].caption = "Extra life"

-- Replace 1-up jingle
local function replaceOneUpJingle(oldname, newname)
	if newname == "_1up" then
		S_StartSound(consoleplayer.mo, sfx_extlif, consoleplayer)
		if not consoleplayer.exiting then P_RestoreMusic(consoleplayer) end
		return true
	end
end

addHook("MusicChange", replaceOneUpJingle)

-- Replace score monitors with 1-up monitors even in record attack
local function noScoreMonitors()
	if modeattacking then
		for mo in mobjs.iterate() do
			if (mo.type == MT_SCORE1K_BOX or mo.type == MT_SCORE10K_BOX)
			and mo.spawnpoint and mo.spawnpoint.type == 409 then
				mo.type = MT_1UP_BOX
			end
		end
	end
end

addHook("MapLoad", noScoreMonitors)

-- Fix double character spawning in marathon mode
local function noDoubleCharacters(p)
	if marathonmode then
		if p.bot and p.playerstate == PST_LIVE
		and (p.mo.skin == players[0].mo.skin) then
			print(p)
			P_SetOrigin(p.mo, 0, 0, FRACUNIT * FRACUNIT)
			P_KillMobj(p.mo)
			-- Dumb fucking hack to eviscerate the bot because
			-- I don't know how to actually remove it
		end
	end
end

addHook("PlayerSpawn", noDoubleCharacters)